#! /usr/bin/env python
# -*- coding: utf-8 -*-


"""
Implementation of an octree data structure
"""

from __future__ import print_function, division, absolute_import


from tpDcc.libs.python import mathlib


class Octree(object):
    """
    An octree data structure partitions 3D space into octants
    """

    def __init__(self, bbox_min, bbox_max):
        """
        Constructor
        :param bbox_min: tuple, contains the minimum X,Y,Z values of the mesh bounding box
        :param bbox_max: tuple, contains the maximum X,Y,Z values of the mesh bounding box
        """

        self._bbox_min = bbox_min
        self._bbox_max = bbox_max
        self._root = OctreeNode(self._bbox_min, self._bbox_max, divisions=0, parent=self)

    # region Properties
    def get_root(self):
        return self._root

    root = property(get_root)
    # endregion


class OctreeNode(object):
    """
    Octant representation
    """

    def __init__(self, bbox_min, bbox_max, divisions=0, parent=None):
        """
        Constructor
        :param bbox_min: tuple, contains the minimum X,Y,Z values of the mesh bounding box
        :param bbox_max: tuple, contains the maximum X,Y,Z values of the mesh bounding box
        :param divisions: int, division level of this octant node (ejp: this node is 2**divisions times smaller than
                               than the original Octree bounding box size
        :param parent: Octree, octree parent for this node
        """

        self._bbox_min = bbox_min
        self._bbox_max = bbox_max
        self._parent = parent
        self._divisions = divisions
        self._children = list()
        self._half_values = mathlib.bounding_box_half_values(bbox_min=bbox_min, bbox_max=bbox_max)

    # region Properties
    def get_divisions(self):
        return self._divisions

    def get_children(self):
        return self._children

    def get_half_values(self):
        return self._half_values

    divisions = property(get_divisions)
    children = property(get_children)
    half_values = property(get_half_values)
    # endregion

    def child_containing(self, point):
        """
        Returns the child node containing the given point
        :param point: tuple(X, Y, Z), tuple containing (X,Y,Z) positions of a point
        :return: OctreeNode
        """

        x, y, z = point
        half_x, half_y, half_z = self._half_values

        # Get octant the point is in
        greater_half_x = int(x >= half_x)
        greater_half_y = int(y >= half_y)
        greater_half_z = int(z >= half_z)

        child_index = 4 * greater_half_z + 2 * greater_half_y + greater_half_x
        return self._children[child_index]

    def is_inside(self, point):
        """
        Returns True if the given points lies inside this OctreeNode, False otherwise
        :param point: tuple(X, Y, Z), tuple containing (X,Y,Z) positions of a point
        :return: bool
        """

        x, y, z = point

        in_x = self._bbox_min[0] <= x < self._bbox_max[0]
        in_y = self._bbox_min[1] <= y < self._bbox_max[1]
        in_z = self._bbox_min[2] <= z < self._bbox_max[2]

        return all(in_x, in_y, in_z)

    def subdivide(self):
        """
        Divides an OctreeNode into 8 octants. The subdivided node replaced in node list by the octants
        :return: OctreeNode, octree node we want to divide
        """

        octant_list = list()
        divs = self._divisions + 1
        min_x, min_y, min_z = self._bbox_min
        max_x, max_y, max_z = self._bbox_max
        half_x, half_y, half_z = self._half_values

        oct_0_min = self._bbox_min  # 000
        oct_0_max = (half_x, half_y, half_z)
        octant_list.append(OctreeNode(oct_0_min, oct_0_max, divisions=divs, parent=self))

        oct_1_min = (half_x, min_y, min_z)  # 001
        oct_1_max = (max_x, half_y, half_z)
        octant_list.append(OctreeNode(oct_1_min, oct_1_max, divisions=divs, parent=self))

        oct_2_min = (min_x, half_y, min_z)  # 010
        oct_2_max = (half_x, max_y, half_z)
        octant_list.append(OctreeNode(oct_2_min, oct_2_max, divisions=divs, parent=self))

        oct_3_min = (half_x, half_y, min_z)  # 011
        oct_3_max = (max_x, max_y, half_z)
        octant_list.append(OctreeNode(oct_3_min, oct_3_max, divisions=divs, parent=self))

        oct_4_min = (min_x, min_y, half_z)  # 100
        oct_4_max = (half_x, half_y, max_z)
        octant_list.append(OctreeNode(oct_4_min, oct_4_max, divisions=divs, parent=self))

        oct_5_min = (half_x, min_y, half_z)  # 101
        oct_5_max = (max_x, half_y, max_z)
        octant_list.append(OctreeNode(oct_5_min, oct_5_max, divisions=divs, parent=self))

        oct_6_min = (min_x, half_y, half_z)  # 110
        oct_6_max = (half_x, max_y, max_z)
        octant_list.append(OctreeNode(oct_6_min, oct_6_max, divisions=divs, parent=self))

        oct_7_min = (half_x, half_y, half_z)  # 111
        oct_7_max = self._bbox_max
        octant_list.append(OctreeNode(oct_7_min, oct_7_max, divisions=divs, parent=self))

        # Remove the original node and add the octants
        self._children = octant_list



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to write/read json files
"""


from __future__ import print_function, division, absolute_import

import os
import json

from tpDcc.libs import python


def write_to_file(data, filename, **kwargs):

    """
    Writes data to JSON file
    """

    if '.json' not in filename:
        filename += '.json'

    indent = kwargs.pop('indent', 2)

    try:
        with open(filename, 'w') as json_file:
            json.dump(data, json_file, indent=indent, **kwargs)
    except IOError:
        python.logger.error('Data not saved to file {}'.format(filename))
        return None

    python.logger.info('File correctly saved to: {}'.format(filename))

    return filename


def read_file(filename):

    """
    Get data from JSON file
    """

    if os.stat(filename).st_size == 0:
        return None
    else:
        try:
            with open(filename, 'r') as json_file:
                data = json.load(json_file)
        except Exception as err:
            python.logger.warning('Could not read {0}'.format(filename))
            raise err

    return data



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to write/read (IO) text files
"""


from __future__ import print_function, division, absolute_import


import os
import sys
import stat
import json
import string
import shutil
import getpass
import datetime
import subprocess
from tempfile import mkstemp
from shutil import move


class FileManager(object):
    """
    Base class to deal with file writing and reading
    """

    def __init__(self, file_path, skip_warning=False):
        """
        :param file_path: str, path to the file to work with
        :param skip_warning: bool, Whether to print warnings or not
        """

        self.file_path = file_path
        self.open_file = None

        if not skip_warning:
            self.check_path(warning_text='Path {} is invalid!'.format(file_path))

    def get_open_file(self):
        """
        Returns managed file and opens it
        :return:  str
        """

        return self.open_file()

    def read_file(self):
        """
        Opens managed file to read
        """

        self.check_file(warning_text='File {} is invalid!'.format(self.file_path))
        self.open_file = open(self.file_path, 'r')

    def write_file(self):
        """
        Opens managed file to write data into (removing any previous data)
        """

        self.check_file(warning_text='File {} is invalid!'.format(self.file_path))
        self.open_file = open(self.file_path, 'w')

    def append_file(self):
        """
        Opens managed file to append data to the current file data
        """

        self.check_file(warning_text='File {} is invalid!'.format(self.file_path))
        self.open_file = open(self.file_path, 'a')

    def close_file(self):
        """
        Close managed file
        """

        if self.open_file:
            self.open_file.close()

    def check_folder(self, warning_text=None):
        """
        Check if a folder is an invalid one and raise error if necessary
        :param warning_text: str
        :return: bool
        """

        from tpDcc.libs.python import path

        if not path.is_dir(self.file_path):
            if warning_text is not None:
                raise NameError(str(warning_text))
            return False

        return True

    def check_path(self, warning_text=None):
        """
        Check if a path to a file is invalid and raise error if necessary
        :param warning_text: str
        :return: bool
        """

        from tpDcc.libs.python import path

        dir_name = path.get_dirname(self.file_path)

        if not path.is_dir(dir_name):
            if warning_text is not None:
                raise UserWarning(str(warning_text))
            return False

        return True

    def check_file(self, warning_text=None):
        """
        Check if a file is an invalid one and raise error if necessary
        :param warning_text: str
        :return: bool
        """

        from tpDcc.libs.python import path

        if not path.is_file(self.file_path):
            if warning_text is not None:
                raise UserWarning(str(warning_text))
            return False

        return True


class FileReader(FileManager, object):
    """
    Class to deal with file read operations
    """

    def __init__(self, file_path):
        super(FileReader, self).__init__(file_path=file_path)

    def read(self):
        """
        Read managed file
        :return: list<str>, list of file lines
        """

        self.read_file()
        lines = self._get_lines()
        self.close_file()

        return lines

    def _get_lines(self):
        try:
            lines = self.open_file.read()
        except Exception:
            return []

        return get_text_lines(lines)


class FileWriter(FileManager, object):
    """
    Class to deal with file write operations
    """

    def __init__(self, file_path):
        super(FileWriter, self).__init__(file_path=file_path)

        from tpDcc.libs.python import osplatform

        osplatform.get_permission(file_path)
        self.append = False

    def write_file(self):
        """
        Overrides write file. This function creates the file if it does not exists
        If append is True, then append any lines to the file instead or replacing
        """

        if self.append:
            self.append_file()
        else:
            super(FileWriter, self).write_file()

    def set_append(self, append):
        """
        If True, next write operations will append new lines to end of documents otherwise the text
        content will be replaced entirely
        :param append: bool
        """

        self.append = append

    def write_line(self, line):
        """
        Writes a single line to the file and closes managed file after write operation
        :param line: str, line to add to the file
        """

        self.write_file()
        try:
            self.open_file.write('%s\n' % line)
        except Exception:
            pass
        self.close_file()

    def write_json(self, dict_data):
        """
        Writes given JSON data (dict) into file
        :param dict_data: dict
        """

        self.write_file()
        try:
            json.dump(dict_data, self.open_file, indent=4, sort_keys=True)
        except Exception:
            pass
        self.close_file()

    def write(self, lines, last_line_empty=True):
        """
        Write the given list of lines to the managed file
        :param lines: list<str>, list of lines to write to the managed file
        :param last_line_empty: bool, whether or not to add a line after the last line
        """

        self.write_file()

        try:
            inc = 0
            for line in lines:
                if inc == len(lines) - 1 and not last_line_empty:
                    self.open_file.write(str('%s' % line))
                    break
                self.open_file.write(str('%s\n' % line))
                inc += 1
        except Exception:
            print('Could not write to file {}'.format(self.file_path))

        self.close_file()


class FileVersion(object):
    """
    Utility class to version files or folders
    """

    def __init__(self, file_path):

        from tpDcc.libs.python import path

        self.file_path = file_path
        if file_path:
            self.filename = path.get_basename(directory=self.file_path)
            self._path = path.get_dirname(file_path)
            self._version_folder_name = '__version__'
            self._version_name = 'version'
            self._version_folder = None
            self.comment_file = None
            self.updated_old = False

    def get_version_name(self):
        return self._version_name

    def set_version_name(self, version_name):
        self._version_name = version_name

    def get_version_folder_name(self):
        return self._version_folder_name

    def set_version_folder_name(self, version_folder_name):
        self._version_folder_name = version_folder_name

    def get_version_folder(self):
        return self._version_folder

    def set_version_folder(self, version_folder):
        self._version_folder = version_folder

    version_name = property(get_version_name, set_version_name)
    version_folder_name = property(get_version_folder_name, set_version_folder_name)
    version_folder = property(get_version_folder, set_version_folder)

    def has_versions(self):
        """
        Returns whether the version file already has versions folder created or not
        :return: bool
        """

        from tpDcc.libs.python import path

        version_folder = self._get_version_folder()
        if path.is_dir(version_folder):
            return True

    def get_latest_version(self):
        """
        Returns the file path to the latest version
        :return: str
        """

        from tpDcc.libs.python import path

        versions = self.get_versions()
        latest_version = versions[-1]

        return path.join_path(self.file_path, '{0}/{1}'.format(self.version_folder_name, latest_version))

    def get_versions(self, return_version_numbers_also=False):
        """
        Get file paths of all version
        :param return_version_numbers_also: Whether the number of the versions should be returned also
        :return: list
        """

        from tpDcc.libs.python import folder, sort

        version_folder = self._get_version_folder()
        files = folder.get_files_and_folders(directory=version_folder)
        if not files:
            return None

        number_list = list()
        pass_files = list()

        for file_path in files:
            if not file_path.startswith(self.version_name):
                continue
            split_name = file_path.split('.')
            if not len(split_name) == 2:
                continue

            number = int(split_name[1])
            number_list.append(number)
            pass_files.append(file_path)

        if not pass_files:
            return

        quick_sort = sort.QuickNumbersListSort(list_of_numbers=number_list)
        quick_sort.set_follower_list(pass_files)
        pass_files = quick_sort.run()

        pass_dict = dict()
        for i in range(len(number_list)):
            pass_dict[pass_files[0][i]] = pass_files[1][i]

        if not return_version_numbers_also:
            return pass_dict
        else:
            return pass_dict, pass_files[0]

    def get_version_numbers(self):
        """
        Return file version numbers of all versions
        :return: list<int>, list of version numbers
        """

        from tpDcc.libs.python import folder

        version_folder = self._get_version_folder()
        files = folder.get_files_and_folders(directory=version_folder)
        if not files:
            return

        number_list = list()
        for file_path in files:
            if not file_path.startswith(self.version_name):
                continue
            split_name = file_path.split('.')
            if not len(split_name) == 2:
                continue
            num = int(split_name[1])
            number_list.append(num)

        return number_list

    def save_comment(self, comment=None, version_file=None):
        """
        Saves a comment to the version file
        :param comment: str, commend to add to the version file
        :param version_file: str, version file
        """

        version = version_file.split('.')
        if version:
            version = version[-1]

        user = getpass.getuser()

        if not comment:
            comment = '-'
        comment.replace('"', '\"')

        comment_file = FileWriter(file_path=self.comment_file)
        comment_file.set_append(True)
        comment_file.write(['version = {0}; comment = "{1}"; user = "{2}"'.format(version, comment, user)])
        comment_file.close_file()

    def save(self, comment=None):
        """
        Saves a new version file
        :param comment: str
        :return: str, new version file name
        """

        from tpDcc.libs.python import folder, path

        if not comment:
            comment = '-'
        comment.replace('\n', '   ')
        comment.replace('\r', '   ')

        self._create_version_folder()
        self._create_comment_file()

        unique_file_name = self._increment_version_file_name()

        if path.is_dir(self.file_path):
            folder.copy_folder(directory=self.file_path, directory_destination=unique_file_name)
        elif path.is_file(self.file_path):
            copy_file(file_path=self.file_path, file_path_destination=unique_file_name)

        self.save_comment(comment=comment, version_file=unique_file_name)

        return unique_file_name

    def get_version_data(self, version_number):
        """
        Returns the version data (comment and user) of the given version number
        :param version_number: int, version number
        :return: list<str, str>, tuple with comment and user of the given version
        """

        from tpDcc.libs.python import path

        file_path = self._get_comment_path()
        if not file_path:
            return None, None

        if path.is_file(file_path):
            read = FileReader(file_path=file_path)
            lines = read.read()

            version = None
            comment = None
            user = None

            for line in lines:
                start_index = line.find('"')
                if start_index > -1:
                    end_index = line.find(';')
                    sub_part = line[start_index + 1:end_index]
                    sub_part = sub_part.replace('"', '\\"')
                    line = line[:start_index + 1] + sub_part + line[end_index:]

                try:
                    exec(line)
                except Exception:
                    pass

                if version == version_number:
                    return comment, user

        return None, None

    def get_all_versions_data(self):
        """
        Returns all the version data (comment, user, file_size, modified and version_file) of all the versions
        :return: list<str, str, str, str, str, str>, tuple version, comment, user, file_size, modified, file_version
        """

        from tpDcc.libs.python import path

        versions = self.get_versions(return_version_numbers_also=True)
        if not versions:
            return
        else:
            version_paths = versions[0]
            version_numbers = versions[1]

        file_path = self._get_comment_path()
        if not file_path:
            return []

        datas = list()
        if path.is_file(file_path):
            read = FileReader(file_path)
            lines = read.read()
            for line in lines:
                line_info_dict = dict()
                version = None
                comment = None
                user = None
                file_size = None
                modified = None

                split_line = line.split(';')
                for sub_line in split_line:
                    assigment = sub_line.split('=')
                    if assigment and assigment[0]:
                        name = assigment[0].strip()
                        value = assigment[1].strip()
                        line_info_dict[name] = value

                # Version
                if 'version' not in line_info_dict:
                    continue
                version = int(line_info_dict['version'])
                if version not in version_numbers:
                    continue

                # Comment
                if 'comment' in line_info_dict:
                    comment = line_info_dict['comment']
                    comment = comment[1:-1]

                # User
                if 'user' in line_info_dict:
                    user = line_info_dict['user']
                    user = user[1:-1]

                # Version File
                version_file = version_paths[(version)]
                version_file = path.join_path(self.file_path, '{0}/{1}'.format(self.version_folder_name, version_file))

                # File Size
                file_size = get_file_size(file_path=version_file)

                # Modified
                modified = get_last_modified_date(file_path=version_file)

                datas.append([version, comment, user, file_size, modified, version_file])

        return datas

    def get_version_path(self, version_number):
        """
        Returns the path to the given version number
        :param version_number: int, version number
        :return: str, path to the version
        """

        return self._get_version_path(version_number=version_number)

    def get_version_comment(self, version_number):
        """
        Returns the comment of the given version number
        :param version_number: int, version number
        :return:  str, version_number comment
        """

        comment, user = self.get_version_data(version_number)
        return comment

    def _get_version_folder(self):
        from tpDcc.libs.python import path

        if path.is_file(self.file_path):
            dir_name = path.get_dirname(self.file_path)
            version_path = path.join_path(dir_name, self._version_folder_name)
        else:
            version_path = path.join_path(self.file_path, self._version_folder_name)

        return version_path

    def _get_version_path(self, version_number):
        from tpDcc.libs.python import path
        return path.join_path(self._get_version_folder(), self._version_name + '.' + str(version_number))

    def _get_version_number(self, file_path):
        from tpDcc.libs.python import name
        version_number = name.get_last_number(input_string=file_path)
        return version_number

    def _get_comment_path(self):
        from tpDcc.libs.python import path
        version_folder = self._get_version_folder()
        file_path = None
        if version_folder:
            file_path = path.join_path(version_folder, 'comments.txt')

        return file_path

    def _create_version_folder(self):
        from tpDcc.libs.python import folder
        self._version_folder = folder.create_folder(name=self._version_folder_name, directory=self._path)

    def _create_comment_file(self):
        self.comment_file = create_file(filename='comments.txt', directory=self._version_folder)

    def _increment_version_file_name(self):
        from tpDcc.libs.python import path
        version_path = path.join_path(self._version_folder, self._version_name + '.1')
        return path.unique_path_name(directory=version_path)


def open_browser(file_path):
    """
    Open the file browser to the path specified
    :param file_path: str, filename with path
    :return:
    """

    # TODO: Only work on Windows, make it cross platform

    from tpDcc.libs.python import osplatform, path

    if not path.is_file(file_path) and not path.is_dir(file_path):
        return

    if osplatform.is_windows():
        os.startfile(file_path)
    elif osplatform.is_linux():
        try:
            opener = 'open' if sys.platform == 'darwin' else 'xdg-open'
            subprocess.call([opener, file_path])
        except Exception as e:
            os.system('gnome-terminal --working-directory={}'.format(file_path))


def create_file(filename, directory, make_unique=False):
    """
    Creates a file
    :param filename: str, name of the new file
    :param directory: str, directory of the new file
    :param make_unique: bool, whether to make the name unique or not
    :return: variant, str || bool, filename with path or False if create file failed
    """

    from tpDcc.libs.python import name, path

    filename = name.clean_file_string(filename)
    full_path = path.join_path(directory, filename)

    if make_unique:
        full_path = path.unique_path_name(full_path)

    try:
        open_file = open(full_path, 'a')
        open_file.close()
    except Exception:
        return False

    return full_path


def copy_file(file_path, file_path_destination):
    """
    Copies the given file to a new given directory
    :param file_path: str, file to copy with full path
    :param file_path_destination: str, destination directory where we want to copy the file into
    :return: str, the new copied path
    """

    shutil.copy2(file_path, file_path_destination)
    return file_path_destination


def move_file(path1, path2):
    """
    Moves the file pointed by path1 under the directory path2
    :param path1: str, file with full path
    :param path2: str, path where path1 should be move into
    :return: bool, Whether the move operation was successfully
    """

    try:
        shutil.move(path1, path2)
    except Exception:
        print('Failed to move {0} to {1}'.format(path1, path2))
        return False

    return True


def delete_file(name, directory=None, show_warning=True):
    """
    Delete the file by name in the directory
    :param name: str, name of the file to delete
    :param directory: str, the directory where the file is stored
    :param show_warning: bool
    :return: str, file path that was deleted
    """

    from tpDcc.libs.python import path, osplatform

    if not directory:
        full_path = name
    else:
        full_path = path.join_path(directory, name)
    if not path.is_file(full_path):
        if show_warning:
            print('File "{}" was not deleted.'.format(full_path))
        return full_path

    try:
        osplatform.get_permission(full_path)
    except Exception:
        pass
    os.remove(full_path)

    return full_path


def rename_file(name, directory, new_name, new_version=False):
    """
    Renames the give nfile in the directory with a new name
    :param name:
    :param directory:
    :param new_name:
    :param new_version:
    :return:
    """

    from tpDcc.libs.python import path

    full_path = path.join_path(directory, name)
    if not path.is_file(full_path):
        return full_path

    new_full_path = path.join_path(directory, new_name)
    if path.is_file(new_full_path):
        print('A file named {} already exists in the directory: {}'.format(new_name, directory))
        return full_path

    os.chmod(full_path, 0o777)
    os.rename(full_path, new_full_path)

    return new_full_path


def write_to_file(file_name, text_to_write):
    """
    Open a file and overwrite its contents of that file with new content
    """

    if os.path.exists(file_name):
        read_only_or_writeable = os.stat(file_name)[0]
        if read_only_or_writeable != stat.S_IWRITE:
            os.chmod(file_name, stat.S_IWRITE)
    file = open(file_name, 'w')
    file.write(text_to_write)
    file.close()


def append_to_file(file_name, text_to_add):
    """
    Open a file and add new context to its existing text
    """

    if os.path.exists(file_name):
        readOnlyOrWriteable = os.stat(file_name)[0]
        if readOnlyOrWriteable != stat.S_IWRITE:
            os.chmod(file_name, stat.S_IWRITE)
    file = open(file_name, 'a')
    file.write(text_to_add)
    file.close()


def replace(file_path, pattern, subst):
    """
    Replaces one string from another string in a given file
    :param file_path: str, path to the file
    :param pattern: search to be replaced
    :param subst: string that will replace the old one
    """

    # Create temp file
    fh, abs_path = mkstemp()
    with os.fdopen(fh, 'w') as new_file:
        with open(file_path) as old_file:
            for line in old_file:
                # TODO: This gives when working with non ascii codecs, solve ethis without try/except
                try:
                    new_file.write(line.replace(pattern, subst))
                except Exception:
                    pass
    # Remove original file
    os.remove(file_path)

    # Move new file
    move(abs_path, file_path)


def remove_extension(file_path):
    """
    Removes extension of the given file path
    For example, C:/test/rig.py will return rig
    :param file_path: str
    :return: str
    """

    split_path = file_path.split('.')
    new_name = file_path
    if len(split_path) > 1:
        new_name = string.join(split_path[:-1], '.')

    return new_name


def is_newer(file1, file2):
    """
    Returns true if file1 is newer than file2.
    :param str file1: first file to compare.
    :param str file2: second file to compare.
    :return: bool, True if file1 is newer or False otherwise
    """

    if not os.path.exists(file1) or not os.path.exists(file2):
        return False

    time1 = os.path.getmtime(file1)
    time2 = os.path.getmtime(file2)
    return time1 > time2


def is_file_in_dir(filename, directory):
    """
    Returns True if the given file name is contained in the given directory
    :param filename: str, file name
    :param directory: str, directory name including path
    :return: bool
    """

    from tpDcc.libs.python import path

    file_path = path.join_path(directory, filename)
    return os.path.isfile(file_path)


def is_same_date(file1, file2):
    """
    Returns True if the given files have the same date
    :param file1: str, file name including path
    :param file2: str, file name including path
    :return: bool
    """

    date1 = os.path.getmtime(file1)
    date2 = os.path.getmtime(file2)
    value = date1 - date2
    if abs(value) < 0.01:
        return True

    return False


def get_file_text(file_path):
    """
    Get the text stored in a file in a unique string (without parsing)
    :param file_path: str
    :return: str
    """

    open_file = open(file_path, 'r')
    lines = ''

    try:
        lines = open_file.read()
        open_file.close()
    except Exception:
        open_file.close()

    return lines


def get_file_size(file_path, round_value=2):
    """
    Returns the size of the given file
    :param file_path: str
    :param round_value: int, value to round size to
    :return: str
    """

    size = os.path.getsize(file_path)
    size_format = round(size * 0.000001, round_value)

    return size_format


def get_size(file_path, round_value=2):
    """
    Return the size of the given directory or file path
    :param file_path: str
    :param round_value: int, value to round size to
    :return: int
    """

    from tpDcc.libs.python import folder, path

    size = 0
    if path.is_dir(file_path):
        size = folder.get_folder_size(file_path, round_value)
    if path.is_file(file_path):
        size = get_file_size(file_path, round_value)

    return size


def get_last_modified_date(file_path):
    """
    Returns the last date given file was modified
    :param file_path: str
    :return: str, formated date and time
    """

    mtime = os.path.getatime(file_path)

    date_value = datetime.datetime.fromtimestamp(mtime)
    year = date_value.year
    month = date_value.month
    day = date_value.day

    hour = str(date_value.hour)
    minute = str(date_value.minute)
    second = str(int(date_value.second))

    if len(hour) == 1:
        hour = '0' + hour
    if len(minute) == 1:
        minute = '0' + minute
    if len(second) == 1:
        second = second + '0'

    return '{0}-{1}-{2}  {3}:{4}:{5}'.format(year, month, day, hour, minute, second)


def get_file_date(file_path):
    """
    Returns date the given file was created
    :param file_path: str
    :return: str
    """

    date_file = 0
    if os.path.isfile(file_path):
        st_file = os.stat(file_path)
        date_file = st_file[stat.ST_MTIME]

    return date_file


def copy_file_date(original_file_path, target_file_path):
    """
    Copies the creation date of one file to another one
    :param original_file_path: str
    :param target_file_path: str
    """

    if os.path.isfile(original_file_path) and os.path.isfile(target_file_path):
        date_file = get_file_date(original_file_path)
        os.utime(target_file_path, (date_file, date_file))


def get_file_lines(file_path):
    """
    Get the text lines from a file
    :param file_path: str, file name of the text to read
    :return: str
    """

    reader = FileReader(file_path=file_path)
    return reader.read()


def get_text_lines(text):
    """
    Get all lines from a text storing each lines as a different item in a list
    :param text: str, text to get lines from
    :return: list<str>
    """

    text = text.replace('\r', '')
    lines = text.split('\n')

    return lines


def write_lines(file_path, lines, append=False):
    """
    Writes a list of text lines to a file. Every entry in the list is a new line
    :param file_path: str, filename and path
    :param lines: list<str>, list of text lines in which each entry is a new line
    :param append: bool, Whether to append the text or replace it
    """

    write_file = FileWriter(file_path=file_path)
    write_file.set_append(append)
    write_file.write(lines)
    write_file.close_file()


def is_same_text_content(file1, file2):
    """
    Returns True if the given text files contains the same text or False otherwise
    :param file1: str, file path to the first text
    :param file2: str, file path to the second text
    :return: bool
    """

    text1 = get_file_text(file_path=file1)
    text2 = get_file_text(file_path=file2)
    if text1 == text2:
        return True

    return False


def get_files(root_directory):
    """
    Returns files found in the given directory
    :param root_directory: str
    :return: list(str)
    """

    found = list()

    files = os.listdir(root_directory)
    for filename in files:
        file_path = os.path.join(root_directory, filename)
        if os.path.isfile(file_path):
            found.append(filename)

    return found


def file_has_info(file_path):
    """
    Check if the given file size is bigger than 1.0 byte
    :param file_path: str, ful file path of the file to check
    :return: bool
    """

    file_stats = os.stat(file_path)
    if file_stats.st_size < 1:
        return False

    return True



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to sort algorithms
"""


from __future__ import print_function, division, absolute_import


class QuickNumbersListSort(object):
    """
    Fast method to sort lists of numbers
    """

    def __init__(self, list_of_numbers):
        self.list_of_numbers = list_of_numbers
        self.follower_list = list()

    def _sort(self, list_of_numbers, follower_list=[]):
        less = list()
        equal = list()
        greater = list()

        if follower_list:
            less_follow = list()
            equal_follow = list()
            greater_follow = list()

        count = len(list_of_numbers)
        if count > 1:
            pivot = list_of_numbers[0]
            for i in range(count):
                value = list_of_numbers[i]
                if follower_list:
                    follower_value = follower_list[i]
                if value < pivot:
                    less.append(value)
                    if follower_list:
                        less_follow.append(follower_value)
                if value == pivot:
                    equal.append(value)
                    if follower_list:
                        equal_follow.append(follower_value)
                if value > pivot:
                    greater.append(value)
                    if follower_list:
                        greater_follow.append(follower_value)

            if not self.follower_list:
                return self._sort(less) + equal + self._sort(greater)
            else:
                less_list_of_numbers, less_follower_list = self._sort(less, less_follow)
                greater_list_of_numbers, greater_follower_list = self._sort(greater, greater_follow)
                list_of_numbers = less_list_of_numbers + equal + greater_list_of_numbers
                follower_list = less_follower_list + equal_follow + greater_follower_list

                return list_of_numbers, follower_list
        else:
            if not self.follower_list:
                return list_of_numbers
            else:
                return list_of_numbers, follower_list

    def set_follower_list(self, list_of_anything):
        """
        This list must match the length of the list given when the class was initialized
        :param list_of_anything: list
        """

        self.follower_list = list_of_anything

    def run(self):
        """
        If not follower list is supplied, return number list sorted else return number list and follower list
        :return:  variant, list || list, list
        """

        if not self.list_of_numbers:
            return
        if self.follower_list and len(self.follower_list) != len(self.list_of_numbers):
            return

        return self._sort(self.list_of_numbers, self.follower_list)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility functions related with attributes
"""

from __future__ import print_function, division, absolute_import


def attr_type(attr_value):
    """
    Determines the attribute type based on the given value
    :param attr_value: variant, attribute value
    :return: str, attribute type
    """

    from tpDcc.libs.python import python

    if python.is_none(attr_value):
        return 'null'
    if python.is_list(attr_value):
        return list_attr_types(attr_value)
    else:
        if python.is_bool(attr_value):
            return 'bool'
        if python.is_string(attr_value):
            return 'str'
        if python.is_number(attr_value):
            if type(attr_value) is float:
                return 'float'
            if type(attr_value) is int:
                return 'int'
    return 'unknown'


def list_attr_types(s):
    """
    Return a string type for the value
    """

    from tpDcc.libs.python import python

    if not python.is_list(s):
        return 'unknown'
    for typ in [str, int, float, bool]:
        if all(isinstance(n, typ) for n in s):
            return '%s%d' % (typ.__name__, len(s))
    if False not in list(set([python.is_number(x) for x in s])):
        return 'float%d' % len(s)
    return 'unknown'


def auto_convert_attr(attr_value):
    """
    Converts a given attribute value to it's given type
    """

    atype = attr_type(attr_value)
    if atype == 'str':
        return str(attr_value)

    if atype == 'bool':
        return bool(attr_value)

    if atype == 'float':
        return float(attr_value)

    if atype == 'int':
        return int(attr_value)

    return attr_value



#!#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility functions related with strings
"""


from __future__ import print_function, division, absolute_import


import re
import os
import random
import logging
from string import ascii_letters

iters = [list, tuple, set, frozenset]


class _hack(tuple):
    pass


iters = _hack(iters)
iters.__doc__ = """
A list of iterable items (like lists, but not strings). Includes whichever
of lists, tuples, sets, and Sets are available in this version of Python.
"""

LOGGER = logging.getLogger()


def _strips(direction, text, remove):
    """
    Strips characters on a certain direction
    :param direction: str, strip direction ("r", "R", "l" or "L")
    :param text: str, text so strip
    :param remove: variant<iter>, elements to remove
    :return:
    """
    if isinstance(remove, iters):
        for subr in remove:
            text = _strips(direction, text, subr)
        return text

    if direction == 'l' or direction == 'L':
        if text.startswith(remove):
            return text[len(remove):]
    elif direction == 'r' or direction == 'R':
        if text.endswith(remove):
            return text[:-len(remove)]
    else:
        raise ValueError("Direction needs to be r or l.")
    return text


def rstrips(text, remove):
    """
    Removes the string `remove` from the right of `text`
    >>> rstrips("foobar", "bar")
    'foo'
    """
    return _strips('r', text, remove)


def lstrips(text, remove):
    """
    Removes the string `remove` from the left of `text`
    >>> lstrips("foobar", "foo")
    'bar'
    >>> lstrips('http://foo.org/', ['http://', 'https://'])
    'foo.org/'
    >>> lstrips('FOOBARBAZ', ['FOO', 'BAR'])
    'BAZ'
    >>> lstrips('FOOBARBAZ', ['BAR', 'FOO'])
    'BARBAZ'
    """
    return _strips('l', text, remove)


def strips(text, remove):
    """
    removes the string `remove` from the both sides of `text`
    >>> strips("foobarfoo", "foo")
    'bar'
    """
    return rstrips(lstrips(text, remove), remove)


def normalize(string):
    """
    Replace all invalid characters with "_"
    :param string: str, string to normalize
    :return: str, normalize string
    """

    string = str(string)

    if re.match('^[0-9]', string):
        string = '_' + string

    return re.sub("[^A-Za-z0-9_-]", "_", str(string))


def remove_invalid_character(string, regex="[^A-Za-z0-9]"):
    """
    Remove all invalid character
    :param string: str, string to normalize
    :return: str, valid string
    """

    return re.sub(regex, "", str(string))


def clean_string(text):
    """
    Returns a cleaned version of a string - removes everything but alphanumeric and characters and dots
    :param text:  str, string to clean
    :return: str, cleaned string
    """

    return re.sub(r'[^a-zA-Z0-9\n\.]', '_', text)


def replace_sharp_with_padding(string, index):
    """
    Replace a list of # symbol with properly padded index (i.e, count_### > count_001)
    :param string: str, string set. Should include '#'
    :param index: int, index to replace
    :return: str, normalized string
    """

    if string.count("#") == 0:
        string += "#"

    digit = str(index)
    while len(digit) < string.count("#"):
        digit = "0" + digit

    return re.sub("#+", digit, string)


def extract(string, start='(', stop=')'):
    """
    Extract the string that is contained between start and stop strings
    :param string: str, string to process
    :param start: str, start string
    :param stop: str, stop string
    :return: str, extracted string
    """

    try:
        return string[string.index(start) + 1:string.index(stop)]
    except Exception:
        return string


def format_path(path):
    """
    Takes a path and format it to forward slashes
    :param path: str
    :return: str
    """

    return os.path.normpath(path).replace('\\', '/').replace('\t', '/t').replace('\n', '/n').replace('\a', '/a')


def format_path_join(path, *paths):
    """
    os.path.join wrapper that returns always a valid Python path
    :param path:  str
    :param paths: str
    :return: str
    """

    return format_path(os.path.join(path, *paths))


def strip_prefix(name, split='_'):
    """
    Strips prefix
    :param name: str, name to strip prefix of
    :param split: str, split character
    :return: str
    """

    if not name.count(split):
        return name

    return split.join(name.split(split)[1:])


def strip_suffix(name, split='_'):
    """
    Returns the portion of name minus the last element separated by the splitter character
    :param name: str, name to strip the suffix from
    :param split: str, split character
    :return: str
    """

    if not name.count(split):
        return name

    return name.replace(split + name.split(split)[-1], '')


def add_prefix(prefix, split, string):
    """
    Adds a prefix to the given string
    :param prefix: str, prefix to add to the string
    :param split: str, split character
    :param string: str, string to add prefix to
    :return: str
    """
    return split.join([prefix, string])


def get_prefix(string, split):
    """
    Returns the prefix of the given string
    :param string: str, string to get prefix of
    :param split: str, split character
    :return: str
    """
    return string.split(split)[0]


def camel_case_to_string(camel_case_string):
    """
    Converts a camel case string to a normal one
    testPath --> test Path
    :param camel_case_string: str
    :return: str
    """

    return re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", camel_case_string)


def string_to_camel_case(string):
    """
    Converts a string to a camel case one
    test path --> TestPath
    :param string: str
    :return: str
    """

    return ''.join(x for x in string.title() if not x.isspace())


def camel_case_to_lower_case_underscore(text):
    """
    Converts camel case string to underscore separate string
    :param text: str, string to convert
    :return: str
    """

    words = list()
    char_pos = 0
    for curr_char_pos, char in enumerate(text):
        if char.isupper() and char_pos < text:
            words.append(text[char_pos:curr_char_pos].lower())
            char_pos = curr_char_pos
    words.append(text[char_pos:].lower())
    return '_'.join(words)


def camel_case_to_title(text):
    """
    Split string by upper case letters and return a nice name
    :param text: str, string to convert
    :return: str
    """

    words = list()
    char_pos = 0
    for curr_char_pos, char in enumerate(text):
        if char.isupper() and char_pos < curr_char_pos:
            words.append(text[char_pos:curr_char_pos].title())
            char_pos = curr_char_pos
    words.append(text[char_pos:].title())
    return ' '.join(words)


def lower_case_underscore_to_camel_case(text):
    """
    Converts string or unicdoe from lower case underscore to camel case
    :param text: str, string to convert
    :return: str
    """

    # NOTE: We use string's class to work on the string to keep its type
    split_string = text.split('_')
    class_ = text.__class__
    return split_string[0] + class_.join('', map(class_.capitalize, split_string[1:]))


def get_trailing_number(input_string, as_string=False, number_count=-1):
    """
    Get the number at the very end of a string. If number not at the end of the string return None
    :param input_string: str
    :param as_string: bool
    :param number_count: int
    :return: varianht, str || None
    """

    if not input_string:
        return None

    number = r'\d+'
    if number_count > 0:
        number = r'\d' * number_count

    group = re.match('([a-zA-Z_0-9]+)(%s$)' % number, input_string)
    if group:
        number = group.group(2)
        if as_string:
            return number
        else:
            return int(number)


def get_string_index(index, padding=2):
    """
    Returns the string equivalent for the given integer index
    :param index: int, the index to get the string equivalent for
    :param padding: int, number of characters for the index string
    :return: str
    """

    str_ind = str(index)
    for i in range(padding - len(str_ind)):
        str_ind = '0' + str_ind

    return str_ind


def get_alpha(value, capital=False):
    """
    Convert an integer value to a character. a-z then double, aa-zz etc.
    @param value: int, Value to get an alphabetic character from
    @param capital: boolean: True if you want to get capital character
    @return: str, Character from an integer
    """

    # Calculate number of characters required
    base_power = base_start = base_end = 0
    while value >= base_end:
        base_power += 1
        base_start = base_end
        base_end += pow(26, base_power)
    base_index = value - base_start

    # Create alpha representation
    alphas = ['a'] * base_power
    for index in range(base_power - 1, -1, -1):
        alphas[index] = chr(97 + (base_index % 26))
        base_index /= 26

    if capital:
        return ''.join(alphas).upper()

    return ''.join(alphas)


def extract_digits_from_end_of_string(input_string):
    """
    Gets digits at the end of a string
    :param input_string: str
    :return: int
    """

    result = re.search(r'(\d+)$', input_string)
    if result is not None:
        return int(result.group(0))


def remove_digits_from_end_of_string(input_string):
    """
    Deletes the numbers at the end of a string
    :param input_string: str
    :return: str
    """

    return re.sub(r'\d+$', '', input_string)


def num_pad(num, length):
    """
    Returns given number with zero's at prefix to match given length
    :param num: int
    :param length: int
    :return: str
    """

    num_str = str(num)
    length_str = str(length)
    num_chars = len(num_str)
    length_chars = len(length_str)
    if num_chars < length_chars:
        diff = length_chars - num_chars
        return '{}{}'.format('0' * diff, num_str)

    return num_str


def rst_to_html(rst):
    """
    Converts given srst strin to HMLT
    :param rst: str
    :return:
    """

    if not rst:
        return ''

    try:
        from docutils import core
    except Exception:
        LOGGER.warning('docutils module is not availble. Impossible to convert RST to HTML ...')
        return rst

    return core.publish_string(rst, writer_name='html').decode('utf-8')


def generate_random_string(num_symbols=5):
    """
    Generates a random string with the given number of characters
    :param num_symbols: int
    :return: str
    """

    result = ''
    for i in range(num_symbols):
        letter = random.choice(ascii_letters)
        result += letter

    return result


def first_letter_lower(input_string):
    """
    Returns a new string with the first letter as lowercase
    :param input_string: str
    :return: str
    """

    return input_string[:1].lower() + input_string[1:]


def first_letter_upper(input_string):
    """
    Returns a new string with the first letter as uppercase
    :param input_string: str
    :return: str
    """

    return input_string[:1].capitalize() + input_string[1:]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to folders
"""


from __future__ import print_function, division, absolute_import


import os
import sys
import errno
import shutil
import logging
import fnmatch
import tempfile
import traceback
import subprocess

from tpDcc.libs import python

LOGGER = logging.getLogger()


def create_folder(name, directory=None, make_unique=False):
    """
    Creates a new folder on the given path and with the given name
    :param name: str, name of the new directory
    :param directory: str, path to the new directory
    :param make_unique: bool, Whether to pad the name with a number to make it unique if the folder is not unique
    :return: variant, str || bool, folder name with path or False if the folder creation failed
    """

    from tpDcc.libs.python import path

    full_path = False

    if directory is None:
        full_path = name

    if not name:
        full_path = directory

    if name and directory:
        full_path = path.join_path(directory, name)

    if make_unique:
        full_path = path.unique_path_name(directory=full_path)

    if not full_path:
        return False

    if path.is_dir(full_path):
        return full_path

    try:
        os.makedirs(full_path)
    except Exception as e:
        return False

    return full_path


def rename_folder(directory, name, make_unique=False):
    """
    Renames given with a new name
    :param directory: str, full path to the diretory we want to rename
    :param name: str, new name of the folder we want to rename
    :param make_unique: bool, Whether to add a number to the folder name to make it unique
    :return: str, path of the renamed folder
    """

    from tpDcc.libs.python import path

    base_name = path.get_basename(directory=directory)
    if base_name == name:
        return

    parent_path = path.get_dirname(directory=directory)
    rename_path = path.join_path(parent_path, name)

    if make_unique:
        rename_path = path.unique_path_name(directory=rename_path)

    if path.is_dir(rename_path) or path.is_file(rename_path):
        return False

    try:
        os.chmod(directory, 0o777)
        message = 'rename: {0} >> {1}'.format(directory, rename_path)
        python.logger.info(message)
        os.rename(directory, rename_path)
    except Exception:
        python.logger.error('{}'.format(traceback.format_exc()))
        return False

    return rename_path


def copy_folder(directory, directory_destination, ignore_patterns=[]):
    """
    Copy the given directory into a new directory
    :param directory: str, directory to copy with full path
    :param directory_destination: str, destination directory
    :param ignore_patterns: list<str>, extensions we want to ignore when copying folder elements
    If ['txt', 'py'] is given all py and text extension files will be ignored during the copy operation
    :return: str, destination directory
    """

    from tpDcc.libs.python import path

    if not path.is_dir(directory=directory):
        return
    if not ignore_patterns:
        shutil.copytree(directory, directory_destination)
    else:
        shutil.copytree(directory, directory_destination, ignore=shutil.ignore_patterns(ignore_patterns))

    return directory_destination


def move_folder(path1, path2):
    """
    Moves the folder pointed by path1 under the directory path2
    :param path1: str, folder with full path
    :param path2: str, path where path1 should be move into
    :return: bool, Whether the move operation was successfully
    """

    try:
        shutil.move(path1, path2)
    except Exception:
        python.logger.warning('Failed to move {0} to {1}'.format(path1, path2))
        return False

    return True


def delete_folder(folder_name, directory=None):
    """
    Deletes the folder by name in the given directory
    :param folder_name: str, name of the folder to delete
    :param directory: str, the directory path where the folder is stored
    :return: str, folder that was deleted with path
    """

    from tpDcc.libs.python import path

    def delete_read_only_error(action, name, exc):
        """
        Helper to delete read only files
        """

        os.chmod(name, 0o777)
        action(name)

    full_path = folder_name
    if directory:
        full_path = path.join_path(directory, folder_name)
    if not path.is_dir(full_path):
        return None

    try:
        shutil.rmtree(full_path, onerror=delete_read_only_error)
    except Exception as exc:
        LOGGER.warning('Could not remove children of path "{}" | {}'.format(full_path, exc))

    return full_path


def clean_folder(directory):
    """
    Removes everything in the given directory
    :param directory: str
    """

    from tpDcc.libs.python import path, fileio, folder

    base_name = path.get_basename(directory=directory)
    dir_name = path.get_dirname(directory=directory)

    if path.is_dir(directory):
        files = folder.get_files(directory)
        for f in files:
            fileio.delete_file(f, directory)

        delete_folder(base_name, dir_name)

    if not path.is_dir(directory):
        create_folder(base_name, dir_name)


def get_folder_size(directory, round_value=2):
    """
    Returns the size of the given folder
    :param directory: str
    :param round_value: int, value to round size to
    :return: str
    """

    from tpDcc.libs.python import path, fileio

    size = 0
    for root, dirs, files in os.walk(directory):
        for name in files:
            size += fileio.get_file_size(path.join_path(root, name), round_value)

    return size


def get_size(file_path, round_value=2):
    """
    Return the size of the given directory or file path
    :param file_path: str
    :param round_value: int, value to round size to
    :return: int
    """

    from tpDcc.libs.python import fileio, path

    size = 0
    if path.is_dir(file_path):
        size = get_folder_size(file_path, round_value)
    if path.is_file(file_path):
        size = fileio.get_file_size(file_path, round_value)

    return size


def get_sub_folders(root_folder, sort=True):
    """
    Return a list with all the sub folders names on a directory
    :param root_folder: str, folder we want to search sub folders for
    :param sort: bool, True if we want sort alphabetically the returned folders or False otherwise
    :return: list<str>, sub folders names
    """

    if not os.path.exists(root_folder):
        raise RuntimeError('Folder {0} does not exists!'.format(root_folder))
    file_names = os.listdir(root_folder)
    result = list()
    for f in file_names:
        if os.path.isdir(os.path.join(os.path.abspath(root_folder), f)):
            result.append(f)
    if sort:
        result.sort()

    return result


def get_folders(root_folder, recursive=False):
    """
    Get folders found in the root folder
    :param root_folder: str, folder we ant to search folders on
    :param recursive: bool, Whether to search in all root folder child folders or not
    :return: list<str>
    """

    from tpDcc.libs.python import path

    found_folders = list()
    if not recursive:
        try:
            found_files = os.listdir(root_folder)
        except Exception:
            return found_folders

        if not found_files:
            return found_folders
        for file_name in found_files:
            folder_name = path.join_path(root_folder, file_name)
            if path.is_dir(folder_name):
                folder_name = os.path.relpath(folder_name, root_folder)
                folder_name = path.clean_path(folder_name)
                found_folders.append(folder_name)
    else:
        try:
            for root, dirs, files in os.walk(root_folder):
                for d in dirs:
                    folder_name = path.join_path(root, d)
                    folder_name = os.path.relpath(folder_name, root_folder)
                    folder_name = path.clean_path(folder_name)
                    found_folders.append(folder_name)
        except Exception:
            return found_folders

    return found_folders


def get_files(root_folder, full_path=False, recursive=False, pattern="*"):
    """
    Returns files found in the given folder
    :param root_folder: str, folder we want to search files on
    :param full_path: bool, if true, full path to the files will be returned otherwise file names will be returned
    :return: list<str>
    """

    from tpDcc.libs.python import path

    if not path.is_dir(root_folder):
        return []

    # TODO: For now pattern only works in recursive searches. Fix it to work on both

    found = list()

    if recursive:
        for dir_path, dir_names, file_names in os.walk(root_folder):
            for file_name in fnmatch.filter(file_names, pattern):
                if full_path:
                    found.append(os.path.join(dir_path, file_name))
                else:
                    found.append(file_name)
    else:
        files = os.listdir(root_folder)
        for f in files:
            file_path = path.join_path(root_folder, f)
            if path.is_file(file_path=file_path):
                if full_path:
                    found.append(file_path)
                else:
                    found.append(f)

    return found


def get_files_and_folders(directory):
    """
    Get files and folders found in the given directory
    :param directory: str, folder we want to get files and folders from
    :return: list<str>
    """

    from tpDcc.libs.python import path

    if not path.is_dir(directory=directory):
        return

    return os.listdir(directory)


def get_files_with_extension(extension, root_directory, full_path=False, recursive=False):
    """
    Returns file in given directory with given extensions
    :param extension: str, extension to find (.py, .data, etc)
    :param root_directory: str, directory path
    :param full_path: bool, Whether to return the file path or just the file names
    :return: list(str)
    """

    found = list()

    if not extension.startswith('.'):
        extension = '.{}'.format(extension)

    if recursive:
        for dir_path, dir_names, file_names in os.walk(root_directory):
            for file_name in file_names:
                filename, found_extension = os.path.splitext(file_name)
                if found_extension == '{}'.format(extension):
                    if not full_path:
                        found.append(file_name)
                    else:
                        found.append(os.path.join(root_directory, file_name))

    objs = os.listdir(root_directory)
    for filename_and_extension in objs:
        filename, found_extension = os.path.splitext(filename_and_extension)
        if found_extension == '{}'.format(extension):
            if not full_path:
                found.append(filename_and_extension)
            else:
                found.append(os.path.join(root_directory, filename_and_extension))

    return found


def get_files_date_sorted(root_directory, extension=None):
    """
    Returns files date sorted found in the given directory
    :param root_directory: str, directory path
    :param extension: str, optional extension to find
    :return: list(str), list of files date sorted in the directory
    """

    def _get_mtime(fld):
        return os.stat(os.path.join(root_directory, fld)).st_mtime

    if not extension:
        files = get_files(root_folder=root_directory)
    else:
        files = get_files_with_extension(extension=extension, root_directory=root_directory)

    return list(sorted(files, key=_get_mtime))


def open_folder(path=None):
    """
    Opens a folder in the explorer in a independent platform way
    If not path is passed the current directory will be opened
    :param path: str, folder path to open
    """

    if path is None:
        path = os.path.curdir
    if sys.platform == 'darwin':
        subprocess.check_call(['open', '--', path])
    elif sys.platform == 'linux2':
        subprocess.Popen(['xdg-open', path])
    elif sys.platform is 'windows' or 'win32' or 'win64':
        new_path = path.replace('/', '\\')
        try:
            subprocess.check_call(['explorer', new_path], shell=False)
        except Exception:
            pass


def get_user_folder(absolute=True):
    """
    Get path to the user folder
    :return: str, path to the user folder
    """

    from tpDcc.libs.python import path

    if absolute:
        return path.clean_path(os.path.abspath(os.path.expanduser('~')))
    else:
        return path.clean_path(os.path.expanduser('~'))


def get_temp_folder():
    """
    Get the path to the temp folder
    :return: str, path to the temp folder
    """

    from tpDcc.libs.python import path

    return path.clean_path(tempfile.gettempdir())


def get_current_working_directory():
    """
    Returns current working directory
    :return: str, path to the current working directory
    """

    return os.getcwd()


def get_folders_from_path(path):
    """
    Gets a list of sub folders in the given path
    :param path: str
    :return: list<str>
    """

    folders = list()
    while True:
        path, folder = os.path.split(path)
        if folder != '':
            folders.append(folder)
        else:
            if path != '':
                folders.append(path)
            break
    folders.reverse()

    return folders


def get_folders_date_sorted(root_folder):
    """
    Returns folder dates sorted found in the given root directory
    :param root_folder: str, directory path
    :return: list(str): list of folder date sorted in the directory
    """

    def _get_mtime(fld):
        return os.stat(os.path.join(root_folder, fld)).st_mtime

    return list(sorted(os.listdir(root_folder), key=_get_mtime))


def ensure_folder_exists(folder_path, persmissions=0o755, place_holder=False):
    """
    Checks that folder given folder exists. If not, folder is created.
    :param folder_path: str, folder path to check or created
    :param persmissions:int, folder permission mode
    :param place_holder: bool, Whether to create place holder text file or not
    :raise OSError: raise OSError if the creation of the folder fails
    """

    if not os.path.exists(folder_path):
        try:
            python.logger.debug('Creating folder {} [{}]'.format(folder_path, persmissions))
            os.makedirs(path, persmissions)
            if place_holder:
                place_path = os.path.join(folder_path, 'placeholder')
                if not os.path.exists(place_path):
                    with open(place_path, 'wt') as fh:
                        fh.write('Automatically generated place holder file')
        except OSError as err:
            if err.errno != errno.EEXIST:
                raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utilities functions for downloading files
"""

from __future__ import print_function, division, absolute_import

import os
import re
import shutil
import zipfile
import logging
try:
    from urllib.request import urlopen, Request
except ImportError:
    from urllib2 import urlopen, Request

numbers = re.compile(r'\d+')

LOGGER = logging.getLogger()


def chunk_report(bytes_so_far, chunk_size, total_size):
    percent = float(bytes_so_far) / total_size
    percent = round(percent * 100, 2)
    LOGGER.info("Downloaded %d of %d bytes (%0.2f%%)\r" % (bytes_so_far, total_size, percent))
    if bytes_so_far >= total_size:
        LOGGER.info('\n')


def chunk_read(response, destination, chunk_size=8192, report_hook=None):
    with open(destination, 'ab') as dst_file:
        total_size = response.info().getheader('Content-Length').strip()
        total_size = int(total_size)
        bytes_so_far = 0
        while 1:
            chunk = response.read(chunk_size)
            dst_file.write(chunk)
            bytes_so_far += len(chunk)
            if not chunk:
                break
            if report_hook:
                report_hook(bytes_so_far=bytes_so_far, chunk_size=chunk_size, total_size=total_size)
    dst_file.close()
    return bytes_so_far


def download_file(filename, destination):
    LOGGER.info('Downloading file {0} to temporary folder -> {1}'.format(os.path.basename(filename), destination))
    try:
        dst_folder = os.path.dirname(destination)
        if not os.path.exists(dst_folder):
            LOGGER.info('Creating downloaded folders ...')
            os.makedirs(dst_folder)

        hdr = {
            'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.11 '
                          '(KHTML, like Gecko) Chrome/23.0.1271.64 Safari/537.11',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Accept-Charset': 'ISO-8859-1,utf-8;q=0.7,*;q=0.3',
            'Accept-Encoding': 'none',
            'Accept-Language': 'en-US,en;q=0.8',
            'Connection': 'keep-alive'}
        req = Request(filename, headers=hdr)
        data = urlopen(req)
        chunk_read(response=data, destination=destination, report_hook=chunk_report)
    except Exception as e:
        raise e

    if os.path.exists(destination):
        LOGGER.info('Files downloaded succesfully!')
        return True
    else:
        LOGGER.info('ERROR: Error when downloading files. Maybe server is down! Please contact TD!')
        return False


def unzip_file(filename, destination, removeFirst=True, removeSubfolders=None):
    LOGGER.info('Unzipping file {} to --> {}'.format(filename, destination))
    try:
        if removeFirst and removeSubfolders:
            LOGGER.info('Removing old installation ...')
            for subfolder in removeSubfolders:
                p = os.path.join(destination, subfolder)
                LOGGER.info('\t{}'.format(p))
                if os.path.exists(p):
                    shutil.rmtree(p)
        if not os.path.exists(destination):
            LOGGER.info('Creating destination folders ...')
            os.makedirs(destination)
        zip_ref = zipfile.ZipFile(filename, 'r')
        zip_ref.extractall(destination)
        zip_ref.close()
        LOGGER.info('Unzip completed!')
    except Exception as e:
        raise e


def get_version(s):
    """
    Look for the last sequence of number(s) in a string and increment
    """

    if numbers.findall(s):
        lastoccr_sre = list(numbers.finditer(s))[-1]
        lastoccr = lastoccr_sre.group()
        return lastoccr
    return None



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Initialization module for tpDcc-libs-python
"""

from __future__ import print_function, division, absolute_import

import os
import sys
import types
import pkgutil
import inspect
import traceback
import importlib
import logging.config
from collections import OrderedDict


class tpPyUtils(object):
    def __init__(self):
        super(tpPyUtils, self).__init__()

        self.loaded_modules = OrderedDict()
        self.reload_modules = list()
        self._module_name = 'tpDcc.libs.python'
        self._module_dir = self.get_module_path()
        self.logger = self.create_logger()

    def get_module_path(self):
        """
        Returns path where tpDcc.libs.python module is stored
        :return: str
        """

        try:
            mod_dir = os.path.dirname(inspect.getframeinfo(inspect.currentframe()).filename)
        except Exception:
            try:
                mod_dir = os.path.dirname(__file__)
            except Exception:
                try:
                    import tpDcc.libs.python
                    mod_dir = tpDcc.libs.python.__path__[0]
                except Exception:
                    return None

        return mod_dir

    def get_data_path(self):
        """
        Returns path where user data should be located
        :return: str
        """

        data_path = os.path.join(os.getenv('APPDATA'), self._module_name)
        if not os.path.isdir(data_path):
            os.makedirs(data_path)

        return data_path

    def create_logger(self):
        """
        Creates and initializes tpDcc.libs.python logger
        """

        from tpDcc.libs.python import log

        log_path = self.get_data_path()
        if not os.path.exists(log_path):
            raise RuntimeError('{} Log Path {} does not exists!'.format(self._module_name, log_path))

        logger = log.create_logger(logger_name=self._module_name, logger_path=log_path).logger
        if '{}_DEV'.format(self._module_name.upper()) in os.environ and os.environ.get('{}_DEV'.format(
                self._module_name.upper())) in ['True', 'true']:
            logger.setLevel(log.LoggerLevel.DEBUG)
        else:
            logger.setLevel(log.LoggerLevel.WARNING)

        return logger

    def import_module(self, module_name):
        """
        Static function used to import a function given its complete name
        :param module_name: str, name of the module we want to import
        """

        try:
            mod = importlib.import_module(module_name)
            self.logger.debug('Imported: {}'.format(mod))
            if mod and isinstance(mod, types.ModuleType):
                return mod
        except (ImportError, AttributeError) as e:
            try:
                self.logger.warning('FAILED IMPORT: {} -> {}'.format(str(module_name), str(e)))
            except Exception:
                self.logger.warning('FAILED IMPORT: {}'.format(module_name))
            self.logger.debug('\t>>>{}'.format(traceback.format_exc()))

    def explore_package(self, module_path, only_packages=False):
        """
        Load module iteratively
        :param module_path: str, name of the module
        :param only_packages: bool, Whether is only packages need to be checked or not
        :return: list<str>, list<str>, list of loaded modules names and list loaded moule paths
        """

        module_names = list()
        module_paths = list()

        def foo(name, only_packages):
            for importer, m_name, is_pkg in pkgutil.iter_modules([name]):
                mod_path = name + '\\' + m_name
                mod_name = '{}.'.format(
                    self._module_name) + os.path.relpath(mod_path, self._module_dir).replace('\\', '.')
                if only_packages:
                    if is_pkg:
                        module_paths.append(mod_path)
                        module_names.append(mod_name)
                else:
                    module_paths.append(mod_path)
                    module_names.append(mod_name)
        foo(name=module_path, only_packages=only_packages)

        return module_names, module_paths

    def import_modules(self, module_path=None):
        """
        Import all the modules of the package
        :param module_path: str, base module name we want to import
        :return:
        """

        if not module_path:
            module_path = self.get_module_path()

        mod_names, mod_paths = self.explore_package(module_path=module_path, only_packages=False)
        for name, _ in zip(mod_names, mod_paths):
            if name not in self.loaded_modules.keys():
                mod = self.import_module(name)
                if mod:
                    if isinstance(mod, types.ModuleType):
                        self.loaded_modules[mod.__name__] = [os.path.dirname(mod.__file__), mod]
                        self.reload_modules.append(mod)

    def import_packages(self, module_path=None, only_packages=False, order=None, skip_packages=None):
        """
        Import all packages of a given omdule
        :param module_path: str, module name
        :param only_packages: bool, Whether to import only packages or not
        :param order: list<str>, list specifying an order for import/reload
        """

        if not skip_packages:
            skip_packages = list()

        if not module_path:
            module_path = self.get_module_path()

        package_names, package_paths = self.explore_package(module_path=module_path, only_packages=only_packages)

        if order is None:
            order = list()

        ordered_names = list()
        ordered_paths = list()
        temp_index = 0
        i = -1
        for o in order:
            for n, p in zip(package_names, package_paths):
                if str(n) == str(o):
                    i += 1
                    temp_index = i
                    ordered_names.append(n)
                    ordered_paths.append(p)
                elif n.endswith(o):
                    ordered_names.insert(temp_index + 1, n)
                    ordered_paths.insert(temp_index + 1, n)
                    temp_index += 1
                elif str(o) in str(n):
                    ordered_names.append(n)
                    ordered_paths.append(p)

        ordered_names.extend(package_names)
        ordered_paths.extend(package_paths)

        names_set = set()
        paths_set = set()
        module_names = [x for x in ordered_names if not (x in names_set or names_set.add(x))]
        module_paths = [x for x in ordered_paths if not (x in paths_set or paths_set.add(x))]

        reloaded_names = list()
        reloaded_paths = list()
        for n, p in zip(package_names, package_paths):
            reloaded_names.append(n)
            reloaded_paths.append(p)

        for name, _ in zip(module_names, module_paths):
            if name not in self.loaded_modules.keys():
                mod = self.import_module(name)
                if mod:
                    if isinstance(mod, types.ModuleType):
                        self.loaded_modules[mod.__name__] = [os.path.dirname(mod.__file__), mod]
                        self.reload_modules.append(mod)

        for name, path in zip(module_names, module_paths):
            order = list()
            can_import = True
            for skip_pkg in skip_packages:
                if name.startswith(skip_pkg):
                    can_import = False
                    break
            if not can_import:
                continue
            if name in self.loaded_modules.keys():
                mod = self.loaded_modules[name][1]
                if hasattr(mod, 'order'):
                    order = mod.order
            self.import_packages(module_path=path, only_packages=False, order=order)

    def reload_all(self):
        """
        Reload all current loaded modules
        """

        for mod in list(sys.modules.keys()):
            if mod in sys.modules:
                if mod == self._module_name:
                    continue
                elif mod.startswith(self._module_name):
                    self.logger.info('Removing module: {}'.format(mod))
                    del sys.modules[mod]


def init(do_reload=True):
    """
    Initializes tpDcc.libs.python package
    :param do_reload: bool, Whether to reload imported tpDcc.libs.python modules or not
    :return: bool
    """

    new_importer = tpPyUtils()

    logger = create_logger()
    register_class('logger', logger)

    new_importer.import_modules()
    new_importer.import_packages(only_packages=True)
    if do_reload:
        new_importer.reload_all()


def create_logger():
    """
    Returns logger of current modue
    """

    logging.config.fileConfig(get_logging_config(), disable_existing_loggers=False)
    logger = logging.getLogger('tpDcc-libs-python')

    return logger


def create_logger_directory():
    """
    Creates artellapipe logger directory
    """

    logger_directory = os.path.normpath(os.path.join(os.path.expanduser('~'), 'tpDcc', 'logs'))
    if not os.path.isdir(logger_directory):
        os.makedirs(logger_directory)


def get_logging_config():
    """
    Returns logging configuration file path
    :return: str
    """

    create_logger_directory()

    return os.path.normpath(os.path.join(os.path.dirname(__file__), '__logging__.ini'))


def register_class(cls_name, cls, is_unique=False):
    """
    This function registers given class
    :param cls_name: str, name of the class we want to register
    :param cls: class, class we want to register
    :param is_unique: bool, Whether if the class should be updated if new class is registered with the same name
    """

    if is_unique:
        if cls_name in sys.modules[__name__].__dict__:
            setattr(sys.modules[__name__], cls_name, getattr(sys.modules[__name__], cls_name))
    else:
        sys.modules[__name__].__dict__[cls_name] = cls



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains interfaces class for different purposes
"""


from __future__ import print_function, division, absolute_import


class ISerializable(object):
    """
    Interface class used to identify serializable/deserializable classes
    """

    def __init__(self):
        super(ISerializable, self).__init__()

    def serialize(self, *args, **kwargs):
        """
        Serializes the current class
        """

        raise NotImplementedError('serialize method of ISerializable is not implemented!')

    def deserialize(self, *args, **kwargs):
        """
        Deserialize the current class
        """

        raise NotImplementedError('deseriailze method of ISerializable is not implemented!')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes and functions related with names
"""

from __future__ import print_function, division, absolute_import

import os
import re
import sys
import string


class FindUniqueString(object):
    """
    Utility class to get unique strings
    """

    def __init__(self, test_string):
        self.test_string = test_string
        self.increment_string = None
        self.padding = 0

    # region Public Functions
    def set_padding(self, padding):
        self.padding = padding

    def get(self):
        return self._search()
    # endregion

    # region Private Functions
    def _get_scope_list(self):
        return []

    def _format_string(self, number):
        if number == 0:
            number = 1

        exp = search_last_number(self.test_string)

        if self.padding:
            number = str(number).zfill(self.padding)

        if exp:
            self.increment_string = '{0}{1}{2}'.format(
                self.test_string[:exp.start()], number, self.test_string[exp.end():])
        else:
            split_dot = self.test_string.split('.')
            if len(split_dot) > 1:
                split_dot[-2] += str(number)
                self.increment_string = string.join(split_dot, '.')
            elif len(split_dot) == 1:
                self.increment_string = '{0}{1}'.format(self.test_string, number)

    def _get_number(self):
        return get_end_number(self.test_string)

    def _search(self):
        number = self._get_number()
        self.increment_string = self.test_string
        unique = False

        while not unique:
            scope = self._get_scope_list()
            if not scope:
                unique = True
                continue
            if self.increment_string not in scope:
                unique = True
                continue
            if self.increment_string in scope:
                if not number:
                    number = 0
                self._format_string(number)
                number += 1
                unique = False
                continue

        return self.increment_string


def strip_name(name):
    """
    Method that strips any |Path and :Namespaces: from a given object DAG path
    Ns:Rig|Ns:Leg|Ns:Test == Test
    :param name: str
    :return: str
    """

    return name.split('|')[-1].split(':')[-1]


def remove_suffix(name):
    """
    Remove suffix from given name string
    @param name: string, given name string to process
    @return: string, name without suffix
    """

    # By convention, we split the names when we find a _ character
    edits = name.split('_')

    # If there is not a _ it means that there is not an underscore and we return the full name
    if len(edits) < 2:
        return name

    # The suffix will be _ + the last split string and the name will be the complete name without the suffix
    suffix = '_' + edits[-1]
    name_no_suffix = name[:-len(suffix)]

    return name_no_suffix


def get_numeric_name(text, names):

    from tpDcc.libs.python import python as utils

    if text in names:
        text = re.sub('\\d*$', '', text)
        names = [n for n in names if n.startswith(text)]
        int_list = []
        for name in names:
            m = re.match('^%s(\\d+)' % text, name)
            if m:
                int_list.append(int(m.group(1)))
            else:
                int_list.append(0)

                int_list.sort()
        missing_int = utils.find_missing_items(int_list)
        if missing_int:
            _id = str(missing_int[0])
        else:
            _id = str(int_list[-1] + 1)
    else:
        _id = ''
    text += _id
    return text


def get_first_number(input_string, as_string=False):
    """
    Returns the first number of the given string
    :param input_string: str
    :param as_string: bool, Whether the found number should be returned as integer or as string
    :return: variant, str || int
    """

    found = re.search('[0-9]+', input_string)
    if not found:
        return None

    number_str = found.group()
    if not number_str:
        return None

    if as_string:
        return number_str
    number = int(number_str)

    return number


def get_last_number(input_string, as_string=False):
    """
    Returns the last number of the given string
    :param input_string: str, string to search for a number
    :param as_string: bool, Whether the found number should be returned as integer or as string
    :return: variant, str || int
    """

    found = search_last_number(input_string)
    if not found:
        return None

    number_str = found.group()
    if not number_str:
        return None

    if as_string:
        return number_str
    number = int(number_str)

    return number


def get_end_number(input_string, as_string=False):
    """
    Get the number at the end of a string
    :param input_string: bool, string to search for a number
    :param as_string: bool, Whether the found number should be returned as integer or as string
    :return: variant, str || int,  number at the end of te string
    """

    found = re.findall(r'\d+', input_string)
    if not found:
        return None

    if type(found) == list:
        found = found[0]

    if as_string:
        return found
    else:
        return int(found)


def get_trailing_number(input_string, as_string=False, number_count=-1):
    """
    Returns the number at the very end of a string. If number not at the end of the string, returns None
    :param input_string: str, string to get trailing number of
    :param as_string: bool, Whether to return the trailing number as an string or an integer
    :param number_count: int, padding trailing count
    :return: variant, str || int
    """

    if not input_string:
        return None

    number = r'\d+'
    if number_count > 0:
        number = r'\d' * number_count

    group = re.match('([a-zA-Z_0-9]+)(%s$)' % number, input_string)
    if group:
        number = group.group(2)
        if as_string:
            return number
        else:
            return int(number)

    return None


def get_trailing_number_data(input_string):
    """
    Returns the trailing number of a string, the name with the number removed and the padding of the number
    :param input_string: str
    :return: tuple(str, str, int)
    """

    m = re.search(r'\d+$', input_string)
    if m:
        num_as_string = m.group()
        name_without_number = input_string[:-len(num_as_string)]
        padding = len(num_as_string)
        return name_without_number, int(num_as_string), padding

    return input_string, None, 0


def get_last_letter(input_string):
    """
    Returns the last letter of the given string
    :param input_string: str, string to search for a letter
    :return: str, last letter in the string
    """

    search = search_last_letter(input_string)
    if not search:
        return None

    found_str = search.group()
    return found_str


def convert_side_name(name):
    """
    Convert a string with underscore to proper side name
    :param name: str, string to convert
    :return: tuple of integer
    """

    if name == "L":
        return "R"
    elif name == "R":
        return "L"
    if name == 'l':
        return 'r'
    if name == 'r':
        return 'l'

    re_pattern = re.compile("_[RLrl][0-9]+_|^[RLrl][0-9]+_|_[RLrl][0-9]+$|_[RLrl]_|^[RLrl]_|_[RLrl]$")

    re_match = re.search(re_pattern, name)
    if re_match:
        instance = re_match.group(0)
        rep = None
        if instance.find("R") != -1:
            rep = instance.replace("R", "L")
        else:
            rep = instance.replace("L", "R")
        if not rep:
            if instance.find('r') != -1:
                rep = instance.replace('r', 'l')
            else:
                rep = instance.replace('l', 'r')

        name = re.sub(re_pattern, rep, name)

    return name


def replace_string(string_value, replace_string, start, end):
    """
    Replaces one string by another
    :param string_value: str, string to replace
    :param replace_string: str, string to replace with
    :param start: int, string index to start replacing from
    :param end: int, string index to end replacing
    :return: str, new string after replacing
    """

    first_part = string_value[:start]
    second_part = string_value[end:]

    return first_part + replace_string + second_part


def replace_string_at_start(line, string_to_replace, replace_string):
    """
    Replaces string at the start of the given line
    :param line: str
    :param string_to_replace: str
    :param replace_string: str
    :return:
    """

    m = re.search('^%s' % string_to_replace, line)
    if not m:
        return

    start = m.start(0)
    end = m.end(0)
    new_line = line[:start] + replace_string + line[end:]

    return new_line


def replace_string_at_end(line, string_to_replace, replace_string):
    """
    Replaces string at the end of the given line
    :param line: str
    :param string_to_replace: str
    :param replace_string: str
    :return:
    """

    m = re.search('%s$' % string_to_replace, line)
    if not m:
        return

    start = m.start(0)
    end = m.end(0)
    new_line = line[:start] + replace_string + line[end:]

    return new_line


def clean_file_string(string):
    """
    Replaces all / and \\ characters by _
    :param string: str, string to clean
    :return: str, cleaned string
    """

    if string == '/':
        return '_'

    string = string.replace('\\', '_')

    return string


def clean_name_string(string_value, clean_chars='_', remove_chars='_'):
    """
    Clean given string by cleaning given clean_char and removing given remove_chars
    :param string_value: str
    :param clean_chars: str
    :param remove_chars: str
    :return: str, cleaned name
    """

    string_value = re.sub('^[^A-Za-z0-9%s]+' % clean_chars, '', string_value)
    string_value = re.sub('[^A-Za-z0-9%s]+$' % clean_chars, '', string_value)
    string_value = re.sub('[^A-Za-z0-9]', remove_chars, string_value)

    if not string_value:
        string_value = remove_chars

    return string_value


def search_first_number(input_string):
    """
    Get the first number in a string
    :param input_string:  string to search for its first number
    :return: int, last number in the string
    """

    regex = re.compile('[0-9]+')
    return regex.search(input_string)


def search_last_number(input_string):
    """
    Get the last number in a string
    :param input_string:  string to search for its lsat number
    :return: int, last number in the string
    """

    regex = re.compile(r'(\d+)(?=(\D+)?$)')
    return regex.search(input_string)


def replace_last_number(input_string, replace_string):
    """
    Replace the last number with the given replace_string
    :param input_string: str, string to search for the last number
    :param replace_string: str, string to replace the last number with
    :return: str, new string after replacing
    """

    replace_string = str(replace_string)
    regex = re.compile(r'(\d+)(?=(\D+)?$)')
    search = regex.search(input_string)
    if not search:
        return input_string + replace_string

    return input_string[:search.start()] + replace_string + input_string[search.end():]


def increment_first_number(input_string, value=1):
    """
    Up the value of the first number by the given value (by default is 1)
    :param input_string: str, string to search for increment its first number
    :return: str, new string after the first number is replaced
    """

    search = search_first_number(input_string)
    if search:
        new_string = '{0}{1}{2}'.format(
            input_string[0:search.start()], int(search.group()) + value, input_string[search.end():])
    else:
        new_string = input_string + '_{}'.format(value)

    return new_string


def increment_last_number(input_string, value=1):
    """
    Up the value of the last number by the given value (by default is 1)
    :param input_string: str, string to search for increment its last number
    :return: str, new string after the last number is replaced
    """

    search = search_last_number(input_string)
    if search:
        new_string = '{0}{1}{2}'.format(
            input_string[0:search.start()], int(search.group()) + value, input_string[search.end():])
    else:
        new_string = input_string + '{}'.format(value)

    return new_string


def search_last_letter(input_string):
    """
    Returns the last letter in a string
    :param input_string: str, string to search for a letter
    :return: str, last letter in the string
    """

    match = re.findall('[_a-zA-Z]+', input_string)
    if match:
        return match[-1][-1]


def format_path(path):
    """
    Takes a path and format it to forward slashes
    :param path: str
    :return: str
    """

    return os.path.normpath(path).replace('\\', '/').replace('\t', '/t').replace('\n', '/n').replace('\a', '/a')


def add_unique_postfix(fn):
    if not os.path.exists(fn):
        return fn

    path, name = os.path.split(fn)
    name, ext = os.path.splitext(name)

    def make_fn(i):
        return os.path.join(path, '%s_%d%s' % (name, i, ext))

    for i in range(2, sys.maxint):
        uni_fn = make_fn(i)
        if not os.path.exists(uni_fn):
            return uni_fn

    return None


def find_unique_name(name, names, inc_format='{name}{count:03}', sanity_count=9999999):
    """
    Finds a unique name in a given set of names
    :param name: str, name to search for in the scene
    :param names: list<str>, set of strings to check for a unique name
    :param inc_format: str, used to increment the name
    :param sanity_count: int, used to prevent infinite search loop. Increment if needed (default=9999999)
    """

    count = 0
    ret = name
    while ret in names:
        count += 1
        ret = inc_format.format(name=name, count=count)
        if sanity_count and count > sanity_count:
            raise Exception('Unable to find a unique name in {} tries, try a different format.'.format(sanity_count))

    return ret


def find_special(pattern, string_value, position_string):
    """
    Searchs given regular expressin pattern in the given string
    :param pattern: str, regular expression pattern to search for
    :param string_value: str, string to search in
    :param position_string: str, 'start', 'end', 'first', 'last', 'inside', where the pattern should search
    :return: tuple(int, int): start and end indexes of the found pattern. Returns (None, None) if pattern is not found
    """

    char_count = len(string_value)
    found_iter = re.finditer(pattern, string_value)
    found = list()
    index_start = None
    index_end = None

    for item in found_iter:
        found.append(item)
    if not found:
        return None, None

    if position_string == 'end':
        index_start = found[-1].start()
        index_end = found[-1].end()

        if index_end > char_count or index_end < char_count:
            return None, None

        return index_start, index_end
    elif position_string == 'start':
        index_start = found[0].start()
        index_end = found[0].end()

        if index_start != 0:
            return None, None
        return index_start, index_end
    elif position_string == 'first':
        index_start = found[0].start()
        index_end = found[0].end()

        return index_start, index_end
    elif position_string == 'last':
        index_start = found[-1].start()
        index_end = found[-1].end()

        return index_start, index_end
    elif position_string == 'inside':
        start_index = None
        end_index = None
        for match in found:
            start_index = match.start()
            end_index = match.end()

            if start_index == 0:
                continue
            if end_index > char_count:
                continue

            break

        index_start = start_index
        index_end = end_index

    return index_start, index_end


def pad_number(name):
    """
    Add a number to a name
    :param name: str, name to pad
    :return: str
    """

    number = get_last_number(name)
    if number is None:
        number = 0
    number_string = str(number)
    index = name.rfind(number_string)
    if number < 10:
        number_string = number_string.zfill(2)

    new_name = name[0:index] + number_string + name[index + 1:]

    return new_name


def find_unique_id(ids):
    """
    Returns a unique int ID from given ids iterable, starting from 1
    :param ids: iterable (list, set, tuple)
    :return: int
    """

    if not ids or len(ids) == 0:
        return 1

    ids = sorted(set(ids))
    last_id = min(ids)

    if last_id > 1:
        return 1

    for uid in ids:
        diff = uid - last_id
        if diff > 1:
            return last_id + 1
        last_id = uid
    else:
        return uid + 1


def get_unique_name_from_list(existing_names, name):
    """
    Creates a unique name by iterating over existing_names and extracts the end digits to find a new unique name
    :param existing_names: list(str), list of strings where to search for existing indexes
    :param name: str, name to obtain a unique version from
    :return: str
    """

    from tpDcc.libs.python import strings

    if name not in existing_names:
        return name

    ids = set()

    for existing_name in existing_names:
        digits = strings.extract_digits_from_end_of_string(existing_name)
        if digits:
            ids.add(digits)
    idx = find_unique_id(ids)
    name_no_digits = strings.remove_digits_from_end_of_string(name)

    return name_no_digits + str(idx)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Modules that contains utility functions related with Python
"""


from __future__ import print_function, division, absolute_import

import os
import sys
import imp
import ast
import uuid
import time
import types
import inspect
import traceback
import collections
from itertools import groupby
from operator import itemgetter

from tpDcc.libs.python import strings


class RollbackImporter(object):
    """
    This class is used to restore the loaded modules in a certain time after the rollback is instantiated
    It does by storing the list of loaded Python modules when rollback is initialized and if the rollback
    is uninstalled, the modules loaded after rollback creation will be deleted, forcing Python to reload
    them when the module is next imported
    """

    def __init__(self):
        """
        Creates an instance and installs and store the current imported modules list
        """

        self._prev_modules = set(sys.modules.keys())

    def uninstall(self):
        """
        Removes current loaded modules loaded after RollbackImporter was instantiated
        """

        for mod_name in sys.modules.keys():
            if mod_name not in self._prev_modules:
                del sys.modules[mod_name]


class classproperty(object):
    """
    Simplified way of creating getter and setters in Python
    """

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


def add_to_python_path(path, check=True, insert=True):
    """
    Adds a path to the Python path, only if it is not present in the Python path
    :param path: str, path to add to the Python path
    :param insert: bool
    """

    if not path:
        return False

    if check:
        if not os.path.exists(path):
            return False

    if path in sys.path:
        return False

    if insert:
        sys.path.insert(0, path)
    else:
        sys.path.append(path)

    return True


def add_to_environment(env, new_paths):
    """
    Adds given paths into the given environment variable
    :param env: str
    :param new_paths: list(str)
    """

    paths = [i for i in os.environ.get(env, '').split(os.pathsep) if i]

    for p in new_paths:
        if p not in paths:
            paths.append(p)

    os.environ[env] = os.pathsep.join(paths) + os.pathsep


def load_python_module(module_name, directory):
    """
    Loads a given module name located in the given directory
    NOTE: After loading a module you can access programmatically to all functions and attributes of the module
    :param module_name: str, name of the module we want to load
    :param directory: str, directory path where the module is located
    :return: mod, loaded module
    """

    from tpDcc.libs.python import path, fileio

    if path.is_dir(directory):
        full_path = path.join_path(directory, module_name)
        if path.is_file(full_path):
            split_name = module_name.split('.')
            file_path, path_name, description = imp.find_module(split_name[0], [directory])
            try:
                module = imp.load_module(module_name, file_path, path_name, description)
            except Exception as e:
                file_path.close()
                return traceback.format_exc()
            finally:
                if file_path:
                    file_path.close()

            return module

    return None


def import_python_module(module_name, directory):
    """
    Imports the given module
    :param module_name: str, name of the module
    :param directory: str, path where the module is located
    :return: mod, imported module
    """

    from tpDcc.libs.python import path, fileio

    if not path.is_dir(directory=directory):
        return

    module = None
    full_path = path.join_path(directory, module_name)
    if path.is_file(full_path):
        if directory not in sys.path:
            sys.path.append(directory)

        split_name = module_name.split('.')
        script_name = split_name[0]

        exec('import {}'.format(script_name))
        exec('reload({})'.format(script_name))

        module = eval(script_name)
        sys.path.remove(directory)

    return module


def source_python_module(module_file):
    """
    Sources the python module located in the given directory
    :param module_file: str
    :return: variant, sourced module || None
    """

    try:
        try:
            module_in = open(module_file, 'rb')
            import md5
            return imp.load_source(md5.new(module_file).hexdigest(), module_file, module_in)
        except Exception:
            return traceback.format_exc()
        finally:
            try:
                module_in.close()
            except Exception:
                pass
    except ImportError:
        traceback.print_exc(file=sys.stderr)
        return None


def get_version():
    """
    Return current Python version used
    :return: SemanticVersion, python version
    """

    from tpDcc.libs.python import version

    py_version = sys.version_info
    current_version = version.SemanticVersion(
        major=py_version.major,
        minor=py_version.minor,
        patch=py_version.micro
    )

    return current_version


def is_python2():
    """
    Returns whether or not current version is Python 2
    :return: bool
    """

    return get_version().major == 2


def is_python3():
    """
    Returns whether or not current version is Python 3
    :return: bool
    """

    return get_version().major == 3


def clear_list(list_to_clear):
    """
    Clears given Python list
    :param list_to_clear: list
    """

    if is_python2():
        del list_to_clear[:]
    else:
        list_to_clear.clear()


def list_diff(list1, list2):
    """
    Returns a a list with all the elements of second list that are not contained in the first list
    :param list1: list
    :param list2: list
    :return: list
    """

    return [i for i in list1 if i not in list2]


def list_to_string(list_):
    """
    Returns a string from the given list
    >>> list_to_string(['1,', '2', '3'])
    >>> # 1, 2, 3
    :param list_: list
    :return:str
    """

    list_ = [str(item) for item in list_]
    list_ = str(list_).replace("[", "").replace("]", "")
    list_ = list_.replace("'", "").replace('"', "")

    return list_


def string_to_list(str_):
    """
    Returns a list from the given string
    >>> string_to_list(['1', '2', '3'])
    >>> # ['1', '2', '3']
    :param str_:
    :return: list
    """

    str_ = '["' + str(str_) + '"]'
    str_ = str_.replace(' ', '')
    str_ = str_.replace(',', '","')

    return eval(str_)


def force_list(var):
    """
    Returns the given variable as list
    :param var: variant
    :return: list
    """

    if var is None:
        return []

    if type(var) is not list:
        if type(var) in [tuple]:
            var = list(var)
        else:
            var = [var]

    return var


def force_tuple(var):
    """
    Returns the given variable as tuple
    :param var: variant
    :return: list
    """

    if type(var) is not tuple:
        var = tuple(var)

    return var


def force_sequence(var, sequence_type=list):
    """
    Returns the given variable as sequence
    If the given variable is list or tuple and sequence_type is different, a conversion will be forced
    :param var: variant
    :return: sequence (tuple || list)
    """

    if type is not list or not tuple:
        sequence_type = list

    if type(var) == list and sequence_type == tuple:
        var = tuple(var)
    if type(var) == tuple and sequence_type == list:
        var = list(var)

    if not type(var) == sequence_type:
        return sequence_type(var)

    return var


def rotate_sequence(seq, current):
    """
    Rotates sequence in given index
    :param seq: list
    :param current: int
    :return: list
    """

    n = len(seq)
    for i in range(n):
        yield seq[(i + current) % n]


def index_of(obj, seq):
    """
    Returns the index of the first occurrence of obj in seq
    """

    try:
        if get_version() < 2.6:
            for index in (i for i in range(len(seq)) if seq[i] == obj):
                return index
            else:
                return next((i for i in range(len(seq)) if seq[i] == obj), None)
    except Exception:
        raise


def last_index_of(obj, seq):
    """
    Returns the index of the last occurrence of obj in seq
    """

    try:
        if get_version() < 2.6:
            for index in (i for i in range(len(seq) - 1, -1, -1) if seq[i] == obj):
                return index
        else:
            return next((i for i in range(len(seq) - 1, -1, -1) if seq[i] == obj), None)
    except Exception:
        raise


def zip_dict(*dcts):
    """
    Custom zip for dictionaries
    """

    for i in set(dcts[0]).intersection(*dcts[1:]):
        yield (i,) + tuple(d[i] for d in dcts)


def module_exists(moduleName):
    """
    Checks if a Python module exists (passing module name)
    :param moduleName: str, Name of module to be checked
    """
    try:
        module_info = imp.find_module(moduleName)
        module = imp.load_module(moduleName, *module_info)
        imp.find_module('__init__', module.__path__)
        has_module = True
    except Exception:
        has_module = False
    return has_module


def find_missing_items(intList):
    """
    Returns missing integers numbers from a list of integers
    :param intList: list(int), list of integers
    :return: list(int),  sorted list of missing integers numbers of the original list
    """

    original_set = set(intList)
    smallest_item = min(original_set)
    largest_item = max(original_set)
    full_set = set(range(smallest_item, largest_item + 1))
    return sorted(list(full_set - original_set))


def safe_code_exec(cmd, env=dict()):
    """
    Safely execute code with specified dictionary
    """

    try:
        exec(cmd in env)
    except Exception:
        raise (sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2])


def remove_dupes(iterable):
    """
    Removes duplicate items from iterable object preserving original iterable order
    :param iterable: iterable
    :return: iterable
    """

    unique = set()
    new_iter = iterable.__class__()
    for item in iterable:
        if item not in unique:
            new_iter.append(item)
        unique.add(item)
    return new_iter


def iter_by(iterable, count):
    """
    Returns a generator which will yield certain elements of the iterable. This elements are defined
    by the count steps
    for obj in iter_by(range(7), 3):
        print(obj)
    [0, 1, 2]
    [3, 4, 5]
    [6]
    :param iterable: iterable
    :param count: int, num of
    :return: generator
    """

    i = iter(iterable)
    while True:
        to_yield = list()
        try:
            for n in range(count):
                to_yield.append(i.next())
                yield to_yield
        except StopIteration:
            if to_yield:
                yield to_yield
            break


def enum(*sequential, **named):
    """
    Enum implementation that supports automatic generation and also supports converting the values
    of the enum back to names
    >>> nums = enum('ZERO', 'ONE', THREE='three')
    >>> nums.ZERO
    # 0
    >>> nums.reverse_mapping['three']
    # 'THREE'
    """

    enums = dict(zip(sequential, range(len(sequential))), **named)
    reverse = dict((value, key) for key, value in enums.items())
    enums['reverse_mapping'] = reverse
    return type(str('Enum'), (), enums)


def generate_uuid():
    """
    Returns a unique UUID
    :return: str
    """

    return str(uuid.uuid4().upper())


def itersubclasses(cls, _seen=None):
    """
    http://code.activestate.com/recipes/576949-find-all-subclasses-of-a-given-class/
    Iterator to yield full inheritance from a given class, including subclasses. This
    """

    if _seen is None:
        _seen = set()
    try:
        subs = cls.__subclasses__()
    except TypeError:
        subs = cls.__subclasses__(cls)

    for sub in subs:
        if sub not in _seen:
            _seen.add(sub)
            yield sub
            for sub in itersubclasses(sub, _seen):
                yield sub


def get_inheritance_map(class_to_process):
    """
    Return the inheritance mapping of the given class
    :param class_to_process: class
    """

    return inspect.getmro(class_to_process)


def get_instance_user_attributes(cls):
    """
    Returns user attributes defined in a class or instance of a class
    :param cls:
    :return: list
    """

    _def_attrs = dir(type('dummy', (object,), {}))
    return [item for item in inspect.getmembers(cls) if item[0] not in _def_attrs]


def user_message(message, prefix='message'):
    """
    Writes teh given prefix + message in the Python output
    :param message: str
    :param prefix: str
    """
    sys.stdout.write(prefix + ': ' + message + '\n')


def is_none(s):
    """
    Returns True if the given object has None type or False otherwise
    :param s: object
    :return: bool
    """

    return type(s).__name__ == 'NoneType'


def is_string(s):
    """
    Returns True if the given object has None type or False otherwise
    :param s: object
    :return: bool
    """
    return type(s) in [str, unicode]


def is_number(s):
    """
    Returns True if given string is int or float, False otherwise
    :param s: object
    :return: bool
    """
    if is_bool(s):
        return False
    return isinstance(s, int) or isinstance(s, float)


def is_bool(s):
    """
    Returns True if the given object has None type or False otherwise
    :param s: object
    :return: bool
    """
    return isinstance(s, bool) or str(s).lower() in ['true', 'false']


def is_list(s):
    """
    Returns True if the given object has list type or False otherwise
    :param s: object
    :return: bool
    """
    return type(s) in [list, tuple]


def is_dict(s):
    """
    Returns True if the given object has dict type or False otherwise
    :param s: object
    :return: bool
    """
    from collections import OrderedDict
    return type(s) in [dict, OrderedDict]


def return_list_without_duplicates(lst):
    """
    Removes duplicates from a list and return the new ilst
    :param lst: list
    :return: list
    """

    new_list = list()
    for item in lst:
        if item not in new_list:
            new_list.append(item)

    return new_list


def delete_pyc_file(python_script):
    """
    Deletes the .pyc file that corresponds to the given .py file
    :param python_script: str
    """

    from tpDcc.libs.python import path, fileio

    script_name = path.get_basename(python_script)
    if not python_script.endswith('.py'):
        print('WARNING: Could not delete .pyc file for {}!'.format(script_name))
        return

    compile_script = python_script + 'c'
    if path.is_file(compile_script):
        compile_name = path.get_basename(compile_script)
        compile_dir_name = path.get_dirname(compile_script)
        if not compile_name.endswith('.pyc'):
            return

        fileio.delete_file(name=compile_name, directory=compile_dir_name)


def convert_list_to_string(*args):
    """
    Given a list returns a string
    :param args:
    :return: str
    """

    try:
        if args is None:
            return 'None'
        if not args:
            return ''
        new_args = list()
        for arg in args:
            if arg is not None:
                new_args.append(str(arg))
        args = new_args
        if not args:
            return ''
        string_value = strings.join(args)
        string_value = string_value.replace('\n', '\t\n')
        if string_value.endswith('\t\n'):
            string_value = string_value[:-2]

        return string_value
    except Exception:
        raise(RuntimeError)


def get_class_parent_classes(base_class=None, parent_class=None):
    """
    Returns all base class parent classes
    :param base_class: obj, Base class
    :param parent_class: obj, Parent class
    :return: list<str>, parent class names
    """

    base_classes = list()
    cl = parent_class if parent_class is not None else base_class
    for b in cl.__bases__:
        if b.__name__ not in ['object']:
            base_classes.append(b)
            base_classes.extend(get_class_parent_classes(parent_class=b))

    return base_classes


def get_dict_ordered_keys_from_values(d):
    """
    Returns a list of keys from dict, ordered based on their values
    :param d: dict, dictionary to construct list from
    :return: list<variant>
    """

    return [i[0] for i in sorted(list(d.items()), key=lambda k, v: v)]


def get_dict_ordered_values_from_keys(d):
    """
    Returns a list of dictionary values, ordered based on their keys
    :param d: dict, dictionary to construct list from
    :return: list<variant>
    """

    return [d[key] for key in sorted(list(d.keys()))]


def flatten_list(list_to_flatten):
    """
    Returns a flattened version of the given list
    :param list_to_flatten: list, list to flatten
    :return: list
    """

    elems = list()
    for i in list_to_flatten:
        if isinstance(i, types.ListType):
            elems.extend(i)
        else:
            elems.append(i)

    return elems


def string_to_dictionary(string):
    """
    Converts a dictionary string representation into a dictionary
    :param string: str
    :return: dict
    """

    return ast.literal_eval(string)


def path_to_dictionary(path):
    """
    Returns the tree hierarchy of the given path as a Python dictionary
    :param path: str
    :return: dict
    """

    d = {'name': os.path.basename(path)}
    if os.path.isdir(path):
        d['type'] = 'directory'
        d['children'] = [path_to_dictionary(os.path.join(path, x)) for x in os.listdir(path)]
    else:
        d['type'] = 'file'

    return d


def to_3_list(item):
    """
    Converts item into a 3 item list
    :param item: var
    :return: list<var, var, var>
    """

    if not isinstance(item, list):
        item = [item] * 3

    return item


def u_print(msg, **kwargs):
    """
    `print` with encoded unicode.
    `print` unicode may cause UnicodeEncodeError
    or non-readable result when `PYTHONIOENCODING` is not set.
    this will fix it.
    :param msg: unicode, message to print
    :param kwargs: dict
    """

    from six import text_type

    if isinstance(msg, text_type):
        encoding = None
        try:
            encoding = os.getenv('PYTHONIOENCODING', sys.stdout.encoding)
        except AttributeError:
            # `sys.stdout.encoding` may not exists.
            pass
        msg = msg.encode(encoding or 'utf-8', 'replace')
    print(msg, **kwargs)


def dict_merge(dict, merge_dict):
    """
    Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    :param dict: dict onto which the merge is executed
    :param merge_dict: dct merged into dct
    :return: None
    """

    for k, v in merge_dict.iteritems():
        if (k in dict and isinstance(dict[k], dict) and isinstance(merge_dict[k], collections.Mapping)):
            dict_merge(dict[k], merge_dict[k])
        else:
            dict[k] = merge_dict[k]


def current_processor_time():
    """
    Retunrs the current processor time
    :return: float
    """

    if is_python2():
        return time.clock()
    else:
        return time.process_time()


def is_iterable(obj):
    """
    Returns whether or not given Python object is iterable
    :param obj: object
    :return: bool
    """

    try:
        it = iter(obj)
    except TypeError:
        return False

    return True


def group_consecutive_items(list_of_items):
    """
    From a list of non grouped consecutive items, returns a list of grouped consecutive frames
    :param list_of_items: list(int), list of frames. For example, [1, 2, 5, 7, 8, 9, 11, 12, 13, 15]
    :return: list(list(int)), list of frames. For example, [[1, 2,], [5], [7, 8, 9], [11, 12, 13], [15]]
    """

    list_of_frames = list()
    for k, g in groupby(enumerate(list_of_items), lambda i, x: i - x):
        list_of_frames.append(map(itemgetter(1), g))

    return list_of_frames


def from_list_to_nested_dict(input_arg, separator='/'):
    """
    Function that converts a list of strings to a nested dict
    :param input_arg: a list/tuple/set of strings
    :param separator: separator split input string
    :return: list of nested dict
    """

    if not isinstance(input_arg, (list, tuple, set)):
        raise TypeError('Input argument "input_arg" should be a list or tuple or set, but get {}'.format(
            type(input_arg)))
    if not isinstance(separator, (str, unicode)):
        raise TypeError('Input argument "separator" should be str or unicode but get {}'.format(type(separator)))

    result = list()

    for item in input_arg:
        components = item.strip(separator).split(separator)
        component_count = len(components)
        current = result
        for i, comp in enumerate(components):
            action_comp = next((x for x in current if x['value'] == comp), None)
            if action_comp is None:
                action_comp = {'value': comp, 'label': comp, 'children': list()}
                current.append(action_comp)
            current = action_comp['children']
            if i == component_count - 1:
                action_comp.pop('children')

    return result


def float_range(start, stop=None, step=None):
    """
    Function that returns a generator that allow us to do a range using float numbers
    :param start: float
    :param stop: float
    :param step: float
    :return: generator
    """

    start = float(start)
    if stop is None:
        stop = start + 0.0
        start = 0.0
    if step is None:
        step = 1.0

    count = 0
    while True:
        temp = float(start + count * step)
        if step > 0 and temp >= stop:
            break
        elif step < 0 and temp <= stop:
            break
        yield temp
        count += 1



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utility functions related with email
"""

import email
import smtplib
import webbrowser
try:
    import urllib2 as urllib
except ImportError:
    import urllib


class Email(object):
    """
    Base class that represents an email
    """

    def __init__(self, user=None, password=None):
        super(Email, self).__init__()

        self.user = user
        self.password = password
        self.emails_list = list()
        self.server = None

        self._setup_message()

    # region Public Functions
    def set_subject(self, subject):
        self.message['Subject'] = subject

    def add_message(self, text):
        self.message.attach(email.MIMEText(text))

    def send(self, emails_list):
        joined = email.Utils.COMMASPACE.join(emails_list)
        self.message['To'] = joined
        self._setup_server()

        if self.server:
            self.server.sendemail(self.user, emails_list, self.message.as_string())
            self.server.quit()
    # endregion

    # region Private Functions
    def _setup_message(self):
        self.message = email.MIMEMultipart.MIMEMultipart()
        self.message['From'] = self.user
        self.message['Subject'] = ''
        self.message['To'] = list()

    def _setup_server(self):
        self.server = smtplib.SMTP()
        self.server.set_debuglevel(1)
    # endregion


class Gmail(Email, object):
    def __init__(self, user=None, password=None):
        super(Gmail, self).__init__(user=user, password=password)

    # region Override Functions
    def _setup_server(self):
        super(Gmail, self)._setup_server()

        smtp_host = 'smtp.gmail.com'
        smtp_port = 587

        self.server.connect(smtp_host, smtp_port)
        self.server.ehlo()
        self.server.starttls()
        self.server.login(self.user, self.password)
    # endregion


class LocalHost(Email, object):
    def __init__(self, user):
        super(LocalHost, self).__init__(user=user, password='')

    # region Override Functions
    def _setup_server(self):
        self.server = smtplib.SMTP('localhost')
    # endregion


def open_web(url):
    """
    Open given web URL in user web browser
    :param url: str
    """

    webbrowser.open(url)


def safe_open_url(url):
    """
    Opens given URL in a safe way
    :param url: str
    :return:
    """

    try:
        result = urllib.urlopen(url)
    except urllib.HTTPError as exc:
        raise Exception('{} : {}'.format(exc, exc.url))

    return result



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation to reload modules recursively
"""

from __future__ import print_function, division, absolute_import


import os
import sys
import inspect
import logging
from distutils import sysconfig

from types import ModuleType

SCRIPT_NAME = 'reloader'
DEBUG_LOG_TITLE = '#' + SCRIPT_NAME + ': '

_processed_modules = set()
_reloaded_modules = set()
_remapping_data = list()

_verbosity = 0
_debug_offset = 0
_debug_offset_string = ''

TAB_SIZE = 4

IGNORE_DIRECTORIES = [
    sysconfig.get_python_lib(standard_lib=True).lower(),
    'autodesk\\maya',
    'pythonpackages',
    'pyqt',
    'keyring',
]


LOGGER = logging.getLogger()


def debug_log(*args, **kwargs):
    verbosity = kwargs.get('verbosity', 1)
    title = kwargs.get('title', True)
    if verbosity <= _verbosity:
        LOGGER.debug((DEBUG_LOG_TITLE if title else '') + _debug_offset_string + ''.join(args))


def debug_log_empty():
    debug_log('', verbosity=1, title=True)


def change_dbg_offset(amount):
    global _debug_offset, _debug_offset_string
    _debug_offset += amount
    _debug_offset_string = ('|' + ' ' * (TAB_SIZE - 1)) * _debug_offset


def need_to_be_ignored(moduleFilename):
    # This is original check for builtin path. Not valid if run in Maya.

    # pyStdLib = sysconfig.get_python_lib(standard_lib=True).lower()
    # pySitePkg = sysconfig.get_python_lib().lower()
    # if moduleFilepath.startswith(pyStdLib) and (not moduleFilepath.startswith(pySitePkg)):
    #     continue

    module_dir = os.path.dirname(moduleFilename).lower()
    for s in IGNORE_DIRECTORIES:
        if s in module_dir:
            return True
    return False


class ImportRemapData(object):

    def __init__(
            self,
            dest_module_name,
            dest_attr_name,
            source_module_name,
            source_attr_name,
            attr_type
    ):

        self._dest_module_name = dest_module_name
        self._dest_attr_name = dest_attr_name
        self._source_module_name = source_module_name
        self._source_attr_name = source_attr_name
        self._attr_type = attr_type

    def remap(self):
        source_module = sys.modules[self._source_module_name]

        if self._attr_type == ModuleType:
            newAttrVal = sys.modules[self._source_attr_name]
        else:
            newAttrVal = getattr(source_module, self._source_attr_name)

        destModule = sys.modules[self._dest_module_name]
        setattr(destModule, self._dest_attr_name, newAttrVal)


def reloader(module, remapping=False, verbosity=0):

    global _verbosity
    _verbosity = verbosity

    debug_log_empty()
    debug_log('RELOAD STARTED.', verbosity=1)

    do_reload(module, remapping=remapping)

    if remapping:
        debug_log('remapping process begins...', verbosity=1)
        for remapData in _remapping_data:
            remapData.remap()
        debug_log('remapping done.', verbosity=1)
    else:
        debug_log('remapping skipped.', verbosity=1)

    _processed_modules.clear()
    _reloaded_modules.clear()
    del _remapping_data[:]

    debug_log('RELOAD FINISHED.', verbosity=1)
    debug_log('', verbosity=1, title=False)


def do_reload(module, remapping=False):

    change_dbg_offset(1)

    debug_log_empty()
    debug_log(module.__name__, ': START PROCESSING.', verbosity=1)
    debug_log_empty()

    _processed_modules.add(module)

    attr_names = dir(module)
    attr_count = len(attr_names)

    detected_modules = set()

    dbg_phase_title = module.__name__ + ': attribute analyzing phase: '
    debug_log(dbg_phase_title, 'analyzing ', str(attr_count), ' module attributes.', verbosity=1)

    for attrName in attr_names:

        attr_val = getattr(module, attrName)
        attr_module = inspect.getmodule(attr_val)

        if not attr_module:
            debug_log(
                dbg_phase_title, '"', attrName, '"(attr): cannot get module for attribute. Skipped.', verbosity=2)
            continue

        debug_log(
            dbg_phase_title, '"', attrName, '"(attr): attribute belongs to module "',
            attr_module.__name__, '"', verbosity=2)

        if attr_module.__name__ in sys.builtin_module_names:
            debug_log(dbg_phase_title, '"', attr_module.__name__, '"(module): builtin. Skipped.', verbosity=2)
            continue

        if not hasattr(attr_module, '__file__'):
            debug_log(
                dbg_phase_title, '"', attr_module.__name__, '"(module): no "__file__" attribute. Skipped.', verbosity=2)
            continue

        if attr_module == sys.modules[__name__]:
            debug_log(
                dbg_phase_title, '"', attr_module.__name__, '"(module): ', SCRIPT_NAME,
                ' itself. Skipped.', verbosity=2)
            continue

        if attr_module in _processed_modules:
            debug_log(
                dbg_phase_title, '"', attr_module.__name__, '"(module): already processed. Skipped.', verbosity=2)
            continue

        if need_to_be_ignored(attr_module.__file__.lower()):
            debug_log(dbg_phase_title, '"', attr_module.__name__, '"(module): in ignore list. Skipped.', verbosity=2)
            continue

        detected_modules.add(attr_module)
        debug_log(
            dbg_phase_title, '"', attr_module.__name__, '"(module): added to list of detected modules.', verbosity=2)

        if remapping:

            if not hasattr(attr_val, '__name__'):
                continue

            orig_name = attr_val.__name__

            if attr_module != module:
                debug_log(' '.join((dbg_phase_title + 'preparing mapping data:', module.__name__, attrName, '<-',
                                    attr_module.__name__, orig_name, str(type(attr_val)))), verbosity=2)
                remap_data = ImportRemapData(
                    module.__name__,
                    attrName,
                    attr_module.__name__,
                    orig_name,
                    type(attr_val)
                )
                _remapping_data.append(remap_data)

    if _verbosity > 0:
        dbg_phase_title = module.__name__ + ': detected modules summary: '
        debug_log_empty()
        if detected_modules:
            debug_log(dbg_phase_title, 'following modules must be processed:', verbosity=1)
            for md in detected_modules:
                debug_log(md.__name__, )
        else:
            debug_log(dbg_phase_title, 'no modules to process.', verbosity=1)

    debug_log_empty()

    dbg_phase_title = module.__name__ + ': recursion entrance phase: '
    for detectedModule in detected_modules:
        if detectedModule in _reloaded_modules:
            debug_log(
                dbg_phase_title,
                'module "' + detectedModule.__name__ + '" was already reloaded. Skipping recursion.', verbosity=1)
            continue
        debug_log(
            dbg_phase_title,
            'module "' + detectedModule.__name__ + '" is not analyzed. Entering recursion.', verbosity=1)
        do_reload(detectedModule)

    dbg_phase_title = module.__name__ + ': reload phase: '
    if module.__name__ != '__main__':
        debug_log(dbg_phase_title, 'reloading module "' + module.__name__ + '".', verbosity=1)
        try:
            reload(module)
        except Exception:
            return
        _reloaded_modules.add(module)
    else:
        debug_log(dbg_phase_title, 'cannot reload "__main__" module. Skipped', verbosity=1)

    debug_log_empty()
    debug_log(module.__name__, ': PROCESSING FINISHED.', verbosity=1)
    debug_log_empty()

    change_dbg_offset(-1)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

"""
Module that contains functions to work with win32
"""

import sys

if 'win' in sys.platform:
    import ctypes
    import ctypes.wintypes

GWL_WNDPROC = -4
GWL_HINSTANCE = -6
GWL_HWNDPARENT = -8
GWL_STYLE = -16
GWL_EXSTYLE = -20
GWL_USERDATA = -21
GWL_ID = -12

WS_BORDER = 0x800000
WS_CAPTION = 0xc00000
WS_CHILD = 0x40000000
WS_CHILDWINDOW = 0x40000000
WS_CLIPCHILDREN = 0x2000000
WS_CLIPSIBLINGS = 0x4000000
WS_DISABLED = 0x8000000
WS_DLGFRAME = 0x400000
WS_GROUP = 0x20000
WS_HSCROLL = 0x100000
WS_ICONIC = 0x20000000
WS_MAXIMIZE = 0x1000000
WS_MAXIMIZEBOX = 0x10000
WS_MINIMIZE = 0x20000000
WS_MINIMIZEBOX = 0x20000
WS_OVERLAPPED = 0
WS_OVERLAPPEDWINDOW = 0xcf0000
WS_POPUP = 0x80000000
WS_POPUPWINDOW = 0x80880000
WS_SIZEBOX = 0x40000
WS_SYSMENU = 0x80000
WS_TABSTOP = 0x10000
WS_THICKFRAME = 0x40000
WS_TILED = 0
WS_TILEDWINDOW = 0xcf0000
WS_VISIBLE = 0x10000000
WS_VSCROLL = 0x200000


def to_hwnd(pycobject):
    """
    Convenience method to get a Windows Handle from a PySide WinID
    Based on http://srinikom.github.io/pyside-bz-archive/523.html
    @return A value equivalent to a void* that represents the Windows handle if one exists; None otherwise.
    """

    if type(pycobject) is long:
        # That specific case happen in maya 2017, here, we already have the hwnd so no further manipulation is needed
        return pycobject
    if sys.version_info[0] == 2:
        ctypes.pythonapi.PyCObject_AsVoidPtr.restype = ctypes.c_void_p
        ctypes.pythonapi.PyCObject_AsVoidPtr.argtypes = [ctypes.py_object]
        return ctypes.pythonapi.PyCObject_AsVoidPtr(pycobject)
    elif sys.version_info[0] == 3:
        ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
        ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object]
        return ctypes.pythonapi.PyCapsule_GetPointer(pycobject, None)


def set_owner(hwnd, hwnd_owner):
    """
    Changes the owner window of the given window
    :param hwnd:
    :param hwnd_owner:
    """

    _update_window = ctypes.windll.user32.UpdateWindow

    # WIN32 vs WIN64 - from a macro in winuser.h
    if ctypes.sizeof(ctypes.wintypes.HWND) == ctypes.sizeof(ctypes.c_long):
        _LONG = ctypes.wintypes.LONG
        _set_window_long = ctypes.windll.user32.SetWindowLongW
        _set_window_long.argtypes = [ctypes.wintypes.HWND, ctypes.c_int, ctypes.wintypes.LONG]
        _set_window_long.restype = ctypes.c_void_p
    elif ctypes.sizeof(ctypes.wintypes.HWND) == ctypes.sizeof(ctypes.c_longlong):
        _LONG = ctypes.wintypes.HWND
        _set_window_long = ctypes.windll.user32.SetWindowLongPtrW
        _set_window_long.argtypes = [ctypes.wintypes.HWND, ctypes.c_int, ctypes.wintypes.HWND]
        _set_window_long.restype = _LONG

    last_error = ctypes.set_last_error(0)
    try:
        result = _set_window_long(ctypes.wintypes.HWND(hwnd), ctypes.c_int(GWL_HWNDPARENT), _LONG(hwnd_owner))
    finally:
        last_error = ctypes.set_last_error(last_error)

    if not result and last_error:
        raise ctypes.WinError(last_error)

    _update_window(hwnd_owner)

    return result


def get_reg_key(registry, key, architecture=None):
    """
    Returns a _winreg hkey if found
    :param registry: str, registry to look in. HKEY_LOCAL_MACHINE for example
    :param key: str, key to open 'Software/Ubisoft/Test' for example
    :param architecture: variant, int || None, 32 or 64 bit. If None, default system architecture is used
    :return: _winreg handle object
    """

    import _winreg

    reg_key = None
    a_reg = _winreg.ConnectRegistry(None, getattr(_winreg, registry))
    if architecture == 32:
        sam = _winreg.KEY_WOW64_32KEY
    elif architecture == 64:
        sam = _winreg.KEY_WOW64_64KEY
    else:
        sam = 0
    try:
        reg_key = _winreg.OpenKey(a_reg, key, 0, _winreg.KEY_READ | sam)
    except WindowsError:
        pass

    return reg_key


def list_reg_keys(registry, key, architecture=None):
    """
    Returns a list of child keys as tuples containing:
        - A string that identifies the value name
        - An object that holds the value data, and whose type depends on the underlying registry type
        - An integer that identifies the type of the value data (see table in docs for _winreg.SetValueEx)
    :param registry: str, registry to look in. HKEY_LOCAL_MACHINE for example
    :param key: str, key to open 'Software/Ubisoft/Test' for example
    :param architecture: variant, int || None, 32 or 64 bit. If None, default system architecture is used
    :return: list<tuple>
    """

    import _winreg

    reg_key = get_reg_key(registry=registry, key=key, architecture=architecture)
    ret = list()
    if reg_key:
        i = 0
        while True:
            try:
                ret.append(_winreg.EnumKey(reg_key, i))
                i += 1
            except WindowsError:
                break

    return ret


def list_reg_key_values(registry, key, architecture=None):
    """
    Returns a list of child keys and their values as tuples containing:
        - A string that identifies the value name
        - An object that holds the value data, and whose type depends on the underlying registry type
        - An integer that identifies the type of the value data (see table in docs for _winreg.SetValueEx)
    :param registry: str, registry to look in. HKEY_LOCAL_MACHINE for example
    :param key: str, key to open 'Software/Ubisoft/Test' for example
    :param architecture: variant, int || None, 32 or 64 bit. If None, default system architecture is used
    :return: list<tuple>
    """

    import _winreg

    reg_key = get_reg_key(registry=registry, key=key, architecture=architecture)
    ret = list()
    if reg_key:
        sub_keys, value_count, modified = _winreg.QueryInfoKey(reg_key)
        for i in range(value_count):
            ret.append(_winreg.EnumValue(reg_key, i))

    return ret


def registry_value(registry, key, value_name, architecture=None):
    """
    Retruns the value and type of the given registry key value name
    :param registry: str, registry to look in. HKEY_LOCAL_MACHINE for example
    :param key: str, key to open 'Software/Ubisoft/Test' for example
    :param value_name: str, name of the value to read. To read the 'default' key, pass an empty string
    :param architecture: variant, int || None, 32 or 64 bit. If None, default system architecture is used
    :return: tuple<object, int, value stored in key and registry type for value (see _winreg's Value Types)
    """

    reg_key = get_reg_key(registry, key, architecture=architecture)
    if reg_key:
        import _winreg
        value = _winreg.QueryValueEx(reg_key, value_name)
        _winreg.CloseKey(reg_key)
        return value

    return '', 0


def get_monitors():
    """
    Returns a list of all monitors
    code.activestate.com/recipes/460509-get-the-actual-and-usable-sizes-of-all-the-monitor
    :return:
    """

    result = list()

    CBFUNC = ctypes.WINFUNCTYPE(ctypes.c_int, ctypes.c_ulong, ctypes.c_ulong, ctypes.POINTER(RECT), ctypes.c_double)

    def cb(h_monitor, hdc_monitor, l_prc_monitor, dw_data):
        r = l_prc_monitor.contents
        data = [h_monitor]
        data.append(r.dump())
        result.append(data)

        return 1

    cb_fn = CBFUNC(cb)
    ctypes.windll.user32.EnumDisplayMonitors(0, 0, cb_fn, 0)

    return result


def get_active_monitor_area():
    """
    Returns the active and working area of each monitor
    code.activestate.com/recipes/460509-get-the-actual-and-usable-sizes-of-all-the-monitor
    :return:
    """

    result = list()
    monitors = get_monitors()
    for h_monitor, extents in monitors:
        data = [h_monitor]
        monitor_info = MonitorInfo()
        monitor_info.cbSize = ctypes.sizeof(MonitorInfo)
        monitor_info.rcMonitor = Rect()
        monitor_info.rcWork = Rect()
        ctypes.windll.user32.GetMonitorInfoA(h_monitor, ctypes.byref(monitor_info))
        data.append(monitor_info.rcMonitor.dump())
        data.append(monitor_info.rcWork.dump())
        result.append(data)

    return result


class Rect(ctypes.Structure):
    _fields_ = [
        ('left', ctypes.c_long),
        ('top', ctypes.c_long),
        ('right', ctypes.c_long),
        ('bottom', ctypes.c_long)
    ]

    def dump(self):
        return tuple(map(int, (self.left, self.top, self.right, self.bottom)))


class MonitorInfo(ctypes.Structure):
    _fields_ = [
        ('cbSize', ctypes.c_long),
        ('rcMonitor', Rect),
        ('rcWork', Rect),
        ('dwFlags', ctypes.c_long)
    ]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to string paths
"""


from __future__ import print_function, division, absolute_import

import os
import stat
import string
import shutil
import logging
import tempfile
import traceback
import contextlib

import six

from tpDcc.libs.python import name, folder

SEPARATOR = '/'
BAD_SEPARATOR = '\\'
PATH_SEPARATOR = '//'
SERVER_PREFIX = '\\'
RELATIVE_PATH_PREFIX = './'
BAD_RELATIVE_PATH_PREFIX = '../'
WEB_PREFIX = 'https://'

# We use one separator depending if we are working on Windows (nt) or other operative system
NATIVE_SEPARATOR = (SEPARATOR, BAD_SEPARATOR)[os.name == 'nt']

LOGGER = logging.getLogger()


class FindUniquePath(name.FindUniqueString, object):
    def __init__(self, directory):
        if not directory:
            directory = folder.get_current_working_directory()

        self.parent_path = self._get_parent_path(directory)
        basename = get_basename(directory=directory)

        super(FindUniquePath, self).__init__(basename)

    def _get_scope_list(self):
        return folder.get_files_and_folders(directory=self.parent_path)

    def _search(self):
        name = super(FindUniquePath, self)._search()
        return join_path(self.parent_path, name)

    def _get_parent_path(self, directory):
        return get_dirname(directory)


@contextlib.contextmanager
def cd(new_dir, cleanup=lambda: True):
    prev_dir = os.getcwd()
    os.chdir(os.path.expanduser(new_dir))
    try:
        yield
    finally:
        os.chdir(prev_dir)
        cleanup()


@contextlib.contextmanager
def temp_dir():
    dir_path = tempfile.mkdtemp()

    def cleanup():
        shutil.rmtree(dir_path)
    with cd(dir_path, cleanup):
        yield dir_path


def normalize_path(path):
    """
    Normalizes a path to make sure that path only contains forward slashes
    :param path: str, path to normalize
    :return: str, normalized path
    """

    path = path.replace(BAD_SEPARATOR, SEPARATOR).replace(PATH_SEPARATOR, SEPARATOR)
    path = six.u(str(path))
    return path.rstrip('/')


def normalize_paths(paths):
    """
    Normalize all the given paths into a consistent format
    :param paths: list(str)
    :return: list(str)
    """

    return [normalize_path(path) for path in paths]


def clean_path(path):
    """
    Cleans a path. Useful to resolve problems with slashes
    :param path: str
    :return: str, clean path
    """

    # We convert '~' Unix character to user's home directory
    path = os.path.expanduser(str(path))

    # Remove spaces from path and fixed bad slashes
    path = normalize_path(path.strip())

    # Fix server paths
    is_server_path = path.startswith(SERVER_PREFIX)
    while SERVER_PREFIX in path:
        path = path.replace(SERVER_PREFIX, PATH_SEPARATOR)
    if is_server_path:
        path = PATH_SEPARATOR + path

    # Fix web paths
    if not path.find(WEB_PREFIX) > -1:
        path = path.replace(PATH_SEPARATOR, SEPARATOR)

    return path


def real_path(path):
    """
    Returns the given path removing any symbolic link
    :param path: str
    :return: str
    """

    path = os.path.realpath(path)
    path = os.path.expanduser(path)

    return normalize_path(path)


def join_path(dir1, dir2):
    """
    Appends dir2 to the end of dir1
    :param dir1: str
    :param dir2: str
    :return: str, combined directory path
    """

    if not dir1 or not dir2:
        return None

    dir1 = clean_path(dir1)
    dir2 = clean_path(dir2)

    dir_path = '{0}/{1}'.format(dir1, dir2)
    dir_path = clean_path(dir_path)

    return dir_path


def split_path(path):
    """
    Split the given path into directory, basename and extension
    :param path:
    :return: list(str)
    """

    path = normalize_path(path)
    filename, extension = os.path.splitext(path)

    return os.path.dirname(filename), os.path.basename(filename), extension


def get_relative_path(path, start):
    """
    Gets a relative path from a start path
    :param path: str, path to get relative path
    :param start: str, Start path to calculate the relative path from
    """

    # if os.path.splitext(start)[-1]:
    #     start = clean_path(os.path.dirname(start))
    # rel_path = clean_path(os.path.relpath(path, start))
    #
    # # TODO: Check if this is correct
    # if not rel_path.startswith('../'):
    #     rel_path = './' + rel_path
    #
    # return rel_path

    rpath = start

    for i in range(0, 3):

        rpath = os.path.dirname(rpath)
        token = os.path.relpath(rpath, start)

        rpath = normalize_path(rpath)
        token = normalize_path(token)

        if rpath.endswith("/"):
            rpath = rpath[:-1]

        path = path.replace(rpath, token)

    return path


def get_absolute_path(path, start):
    """
    Gets an absolute path from a start path
    :param path: str, path to get absolute path
    :param start: str, Start path to calculate the absolute path from
    """

    path = path.replace('\\', '/')
    if not os.path.isdir(start):
        start = os.path.dirname(start).replace('\\', '/')
    else:
        start = start.replace('\\', '/')

    return os.path.abspath(os.path.join(start, path)).replace('\\', '/')


def get_absolute_file_paths(root_directory):
    """
    Returns a generator with all absolute paths on a folder (and sub folders)
    :param root_directory: str, directory to start looking
    """

    for root, _, files in os.walk(root_directory):
        for f in files:
            yield os.path.abspath(os.path.join(root, f))


def get_immediate_subdirectories(root_directory):
    """
    Returns a list with intermediate subdirectories of root directory
    :param root_directory: str, directory to start looking
    """

    return [
        os.path.join(
            root_directory, name) for name in os.listdir(root_directory) if os.path.isdir(
            os.path.join(root_directory, name))
    ]


def get_extension(path):
    """
    Returns the exctension of a file path (wihtout the period)
    :param path: str, valid path to a file
    :return: str
    """

    return os.path.splitext(path)[1][1:]


def exists(directory):
    """
    Returns true if the given path exists
    :param directory: str
    :return: bool
    """

    return os.path.exists(directory)


def has_extension(path, file_extension):
    """
    Checks if a given file path has a specific given file extension
    :param path: str, file path
    :param file_extension: str, valid file extension
    :return: bool, True if the extension of the given path matches the given extension, False otherwise
    """

    return True if get_extension(path=path) == file_extension else False


def get_files(root, file_extension=None, recursive=False, full_path=False, stdout=False):
    """
    Get all files from a given directory
    :param root: str, path to get directories from
    :param file_extension: str, file extension of files to search for
    :param recursive: bool, True if the function will search deeper than one level of files
    :param full_path: str, the output of the path will be the full path if True
    :param stdout: print results in Python output if True
    :return: list<str>, list of files
    """

    def out(data):
        for i in data:
            print(i)
            print('Found {0} files'.format(len(data)))

    if len(root):
        root = clean_path(root)
        directories = [d for d in os.listdir(root) if is_dir(root, d)]
        if file_extension:
            if full_path:
                files = [clean_path(os.path.abspath(os.path.join(root, f))) for
                         f in os.listdir(root) if is_file(root, f) and has_extension(f, file_extension)]
            else:
                files = [f for f in os.listdir(root) if is_file(root, f) and has_extension(f, file_extension)]
        else:
            if full_path:
                files = [
                    clean_path(os.path.abspath(os.path.join(root, f))) for f in os.listdir(root) if is_file(root, f)]
            else:
                files = [f for f in os.listdir(root) if is_file(root, f)]

        if len(directories) and recursive:
            more_files = [get_files(os.path.join(root, d), file_extension, recursive, full_path) for d in directories]
            if len(more_files):
                for chunk in more_files:
                    files.extend(chunk)

        output = [clean_path(p) for p in files]
        if stdout:
            out(output)

        return output
    return None


def get_folders_from_path(path):
    """
    Gets a list of sub folders in the given path
    :param path: str
    :return: list<str>
    """

    folders = list()
    while True:
        path, folder = os.path.split(path)
        if folder != '':
            folders.append(folder)
        else:
            if path != '':
                folders.append(path)
            break
    folders.reverse()

    return folders


def get_basename(directory, with_extension=True):
    """
    Get the last part of a directory name
    For example, C:/test/rig.py will return rig.py if with_extension is True of rig if with_extension is False
    :param directory: str
    :param with_extension: bool, Whether to return the file name with extension or not
    :return: variant, str || bool (if fails)
    """

    from tpDcc.libs.python import fileio

    try:
        base_name = os.path.basename(directory)
        if not with_extension:
            base_name = fileio.remove_extension(base_name)

        return base_name
    except Exception:
        return False


def get_dirname(directory):
    """
    Given a directory path, this will return the path above the last child path in the path
    For example, C:/test/rig.py will return C:/test
    :param directory: str,
    :return: variant, str || bool (if fails)
    """

    try:
        return os.path.dirname(directory)
    except Exception:
        return False


def unique_path_name(directory, padding=0):
    """
    Add padding to the given path name if it is not unique
    :param directory: str, diretory name including path
    :param padding: int, where the padding should start
    :return: str, new unique directory with path
    """

    unique_path = FindUniquePath(directory)
    unique_path.set_padding(padding)

    return unique_path.get()


def get_common_path(path1, path2):
    """
    Returns path that is common in both given paths
    :param path1: str
    :param path2: str
    :return: str, common path shared by both given paths
    """

    path1 = clean_path(path1)
    path2 = clean_path(path2)

    split_path1 = path1.split('/')
    split_path2 = path2.split('/')

    first_list = split_path1
    second_list = split_path2

    found = list()
    for i in range(len(first_list)):
        if len(second_list) <= i:
            break
        if first_list[i] == second_list[i]:
            found.append(first_list[i])
        if first_list[i] != second_list[i]:
            break

    found = string.join(found, '/')

    return found


def remove_common_path(path1, path2):
    """
    Removes path that is common in both given paths
    :param path1: str
    :param path2: str
    :return: str, path without the path shared by both given paths
    """

    path1 = clean_path(path1)
    path2 = clean_path(path2)

    split_path1 = path1.split('/')
    split_path2 = path2.split('/')

    skip = True
    new_path = list()

    for i in range(len(split_path2)):
        if skip:
            if len(split_path1) > i:
                if split_path1[i] != split_path2[i]:
                    skip = False
            if (len(split_path1) - 1) < i:
                skip = False
        if not skip:
            new_path.append(split_path2[i])

    new_path = string.join(new_path, '/')

    return new_path


def remove_common_path_at_beginning(path1, path2):
    """
    Removes path that is similar on both given paths at the beginning of both of them
    :param path1: str
    :param path2: str
    :return: str
    """

    value = path2.find(path1)
    sub_part = None

    if value > -1 and value == 0:
        sub_part = path2[len(path1):]
    if sub_part:
        if sub_part.startswith('/'):
            sub_part = sub_part[1:]

    return sub_part


def is_dir(directory, path=None):
    """
    Checks if the given directory is a directory or not
    :param directory: str
    :return: bool
    """

    if not directory:
        return False

    if path is not None:
        directory = join_path(directory, path)

    try:
        mode = os.stat(directory)[stat.ST_MODE]
        if stat.S_ISDIR(mode):
            return True
    except Exception:
        return False

    return False


def is_file(file_path, path=None):
    """
    Checks if the given path is an existing file
    :param file_path: str
    :return: bool
    """

    if not file_path:
        return False

    if path is not None:
        file_path = join_path(file_path, path)

    try:
        mode = os.stat(file_path)[stat.ST_MODE]
        if stat.S_ISREG(mode):
            return True
    except Exception:
        return False

    return False


def move(path1, path2):
    """
    Move the folder or file pointed by path1 under the directory path2
    :param path1: str, file or folder including path
    :param path2: str, path where path1 should move to
    :return: bool, Whether the move operation was successful
    """

    try:
        shutil.move(path1, path2)
    except Exception:
        LOGGER.warning('Failed to move {0} to {1}'.format(path1, path2))
        return False

    return True


def rename(directory, name, make_unique=False):
    """
    Renames given with a new name
    :param directory: str, full path to the diretory we want to rename
    :param name: str, new name of the folder we want to rename
    :param make_unique: bool, Whether to add a number to the folder name to make it unique
    :return: str, path of the renamed folder
    """

    base_name = get_basename(directory=directory)
    if base_name == name:
        return

    parent_path = get_dirname(directory=directory)
    rename_path = join_path(parent_path, name)

    if make_unique:
        rename_path = unique_path_name(directory=rename_path)

    if is_dir(rename_path) or is_file(rename_path):
        return False

    try:
        os.chmod(directory, 0o777)
        message = 'rename: {0} >> {1}'.format(directory, rename_path)
        LOGGER.info(message)
        os.rename(directory, rename_path)
    except Exception:
        LOGGER.error('{}'.format(traceback.format_exc()))
        return False

    return rename_path



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains basic class for Python importers
"""


from __future__ import print_function, division, absolute_import

import os
import sys
import types
import pkgutil
import inspect
import traceback
import importlib
from collections import OrderedDict

from tpDcc.libs.python import decorators, log as log_utils


class Importer(object):
    """
    Base class that allows to import/reload all the modules in a given package and in a given order
    """

    def __init__(self, module_name, module_dir=None, logger=None, debug=False):
        super(Importer, self).__init__()

        if module_dir is None:
            module_dir = self.get_module_path()

        self._module_name = module_name
        self._module_dir = module_dir

        self.loaded_modules = OrderedDict()
        self.reload_modules = list()

        if logger is None:
            self.log, self.logger = self.create_logger(debug=debug)
        else:
            self.log = None
            self.logger = logger

    @decorators.abstractmethod
    def get_module_path(self):
        """
        Returns path where importer is located
        Must be override in class
        """

        raise NotImplementedError('get_module_path() is not implemented!')

    def get_data_path(self):
        """
        Returns path where user data should be located
        :return: str
        """

        data_path = os.path.join(os.getenv('APPDATA'), self._module_name)
        if not os.path.isdir(data_path):
            os.makedirs(data_path)

        return data_path

    def create_logger(self, debug=False):
        """
        Creates and initializes importer logger
        """

        log_path = self.get_data_path()
        if not os.path.exists(log_path):
            raise RuntimeError('{} Log Path {} does not exists!'.format(self._module_name, log_path))

        log = log_utils.create_logger(logger_name=self._module_name, logger_path=log_path)
        logger = log.logger

        debug_env_var = '{}_DEV'.format(self._module_name.upper())
        if debug:
            os.environ[debug_env_var] = 'True'
        # else:
        #     os.environ[debug_env_var] = 'False'
        if debug_env_var in os.environ and os.environ.get(debug_env_var) in ['True', 'true']:
            logger.setLevel(log_utils.LoggerLevel.DEBUG)
        else:
            logger.setLevel(log_utils.LoggerLevel.WARNING)

        return log, logger

    def import_module(self, module_name):
        """
        Static function used to import a function given its complete name
        :param module_name: str, name of the module we want to import
        """

        try:
            mod = importlib.import_module(module_name)
            self.logger.debug('Imported: {}'.format(mod))
            if mod and isinstance(mod, types.ModuleType):
                return mod
        except Exception as e:
            try:
                self.logger.warning('FAILED IMPORT: {} -> {}'.format(str(module_name), str(traceback.format_exc())))
            except Exception:
                self.logger.warning('FAILED IMPORT: {}'.format(module_name))
            self.logger.debug('\t>>>{}'.format(traceback.format_exc()))

    def explore_package(self, module_path, only_packages=False):
        """
        Load module iteratively
        :param module_path: str, name of the module
        :param only_packages: bool, Whether is only packages need to be checked or not
        :return: list<str>, list<str>, list of loaded modules names and list loaded moule paths
        """

        module_names = list()
        module_paths = list()

        def foo(name, only_packages):
            for importer, m_name, is_pkg in pkgutil.iter_modules([name]):
                mod_path = name + '\\' + m_name
                mod_name = '{}.'.format(
                    self._module_name) + os.path.relpath(mod_path, self._module_dir).replace('\\', '.')
                if only_packages:
                    if is_pkg:
                        module_paths.append(mod_path)
                        module_names.append(mod_name)
                else:
                    module_paths.append(mod_path)
                    module_names.append(mod_name)
        foo(name=module_path, only_packages=only_packages)

        return module_names, module_paths

    def import_modules(self, module_path=None, skip_modules=None):
        """
        Import all the modules of the package
        :param module_path: str, base module name we want to import
        :param skip_modules: list(str)
        :return:
        """

        if skip_modules is None:
            skip_modules = list()

        if not module_path:
            module_path = self.get_module_path()

        mod_names, mod_paths = self.explore_package(module_path=module_path, only_packages=False)
        for name, _ in zip(mod_names, mod_paths):
            if name not in self.loaded_modules.keys():
                if name in skip_modules:
                    continue
                mod = self.import_module(name)
                if mod:
                    if isinstance(mod, types.ModuleType):
                        self.loaded_modules[mod.__name__] = [os.path.dirname(mod.__file__), mod]
                        self.reload_modules.append(mod)

    def import_packages(self, module_path=None, only_packages=False, order=None, skip_modules=None):
        """
        Import all packages of a given omdule
        :param module_path: str, module name
        :param only_packages: bool, Whether to import only packages or not
        :param order: list<str>, list specifying an order for import/reload
        """

        if skip_modules is None:
            skip_modules = list()

        if not module_path:
            module_path = self.get_module_path()

        package_names, package_paths = self.explore_package(module_path=module_path, only_packages=only_packages)

        if order is None:
            order = list()

        ordered_names = list()
        ordered_paths = list()
        temp_index = 0
        i = -1
        for o in order:
            for n, p in zip(package_names, package_paths):
                if str(n) == str(o):
                    i += 1
                    temp_index = i
                    ordered_names.append(n)
                    ordered_paths.append(p)
                elif n.endswith(o):
                    ordered_names.insert(temp_index + 1, n)
                    ordered_paths.insert(temp_index + 1, n)
                    temp_index += 1
                # elif str(o) in str(n):
                #     ordered_names.append(n)
                #     ordered_paths.append(p)

        ordered_names.extend(package_names)
        ordered_paths.extend(package_paths)

        names_set = set()
        paths_set = set()
        module_names = [x for x in ordered_names if not (x in names_set or names_set.add(x))]
        module_paths = [x for x in ordered_paths if not (x in paths_set or paths_set.add(x))]

        reloaded_names = list()
        reloaded_paths = list()
        for n, p in zip(package_names, package_paths):
            reloaded_names.append(n)
            reloaded_paths.append(p)

        for name, _ in zip(module_names, module_paths):
            if name not in self.loaded_modules.keys():
                skip = False
                for skip_module in skip_modules:
                    if skip_module == name or name.startswith(skip_module):
                        skip = True
                        break
                if skip:
                    continue
                mod = self.import_module(name)
                if mod:
                    if isinstance(mod, types.ModuleType):
                        self.loaded_modules[mod.__name__] = [os.path.dirname(mod.__file__), mod]
                        self.reload_modules.append(mod)

        for name, path in zip(module_names, module_paths):
            order = list()
            if name in self.loaded_modules.keys():
                mod = self.loaded_modules[name][1]
                if hasattr(mod, 'order'):
                    order = mod.order
            self.import_packages(module_path=path, only_packages=False, order=order, skip_modules=skip_modules)

    def reload_all(self):
        """
        Reload all current loaded modules
        """

        for mod in self.reload_modules:
            if not hasattr(mod, 'no_reload'):
                self.logger.debug('Reloading: {}'.format(mod.__name__))
                reload(mod)
            else:
                self.logger.debug('Avoiding reload of: {}'.format(mod.__name__))
        self.logger.debug('{} reloaded successfully!'.format(self._module_name))


class SimpleImporter(object):
    """
    Simple importer in which module order cannot be defined useful in some scenarios
    """

    def __init__(self, module_name, logger=None):
        super(SimpleImporter, self).__init__()

        self._module_name = module_name

        if logger is None:
            self.log, self.logger = self.create_logger()
        else:
            self.log = None
            self.logger = logger

    @decorators.abstractmethod
    def get_module_path(self):
        """
        Returns path where importer is located
        Must be override in class
        """

        raise NotImplementedError('get_module_path() is not implemented!')

    def get_data_path(self):
        """
        Returns path where user data should be located
        :return: str
        """

        data_path = os.path.join(os.getenv('APPDATA'), self._module_name)
        if not os.path.isdir(data_path):
            os.makedirs(data_path)

        return data_path

    def create_logger(self):
        """
        Creates and initializes importer logger
        """

        log_path = self.get_data_path()
        if not os.path.exists(log_path):
            raise RuntimeError('{} Log Path {} does not exists!'.format(self._module_name, log_path))

        log = log_utils.create_logger(logger_name=self._module_name, logger_path=log_path)
        logger = log.logger

        if '{}_DEV'.format(self._module_name.upper()) in os.environ and os.environ.get('{}_DEV'.format(
                self._module_name.upper())) in ['True', 'true']:
            logger.setLevel(log_utils.LoggerLevel.DEBUG)
        else:
            logger.setLevel(log_utils.LoggerLevel.WARNING)

        return log, logger

    def import_modules(self):
        """
        This function imports all the modules located in the given importer directory
        """

        import inspect
        scripts_dir = self.get_module_path()
        for key, module in sys.modules.items():
            try:
                module_path = inspect.getfile(module)
            except TypeError:
                continue
            if module_path == scripts_dir:
                continue
            if module_path.startswith(scripts_dir):
                try:
                    importlib.import_module(module.__name__)
                except Exception as e:
                    self.logger.error('{} | {}'.format(e, traceback.format_exc()))

    def reload_all(self):
        """
        Reloads all the modules of the given importer module name
        """

        for mod in sys.modules.keys():
            if mod in sys.modules and mod.startswith(self._module_name):
                del sys.modules[mod]
        self.import_modules()


def init_importer(importer_class, do_import=False, do_reload=True, debug=False):
    """
    Initializes importer
    :param importer_class:
    :param do_import: bool
    :param do_reload: bool
    :param debug: bool
    :return:
    """

    if inspect.isclass(importer_class):
        new_importer = importer_class(debug=debug)
    else:
        new_importer = importer_class

    if do_reload:
        new_importer.reload_all()

    if do_import:
        new_importer.import_modules()
        if hasattr(importer_class, 'import_packages'):
            new_importer.import_packages(only_packages=True)

    return new_importer
